import typing

from typing_extensions import Literal
from pydantic import field_validator, BaseModel, Field, PlainSerializer
from datetime import datetime

from opentrons_shared_data.pipette.types import LabwareUri

from opentrons.types import Point
from opentrons.calibration_storage import types

DatetimeType = typing.Annotated[
    datetime,
    PlainSerializer(lambda x: x.isoformat(), when_used="json"),
]


class CalibrationStatus(BaseModel):
    markedBad: bool = False
    source: typing.Optional[types.SourceType] = None
    markedAt: typing.Optional[DatetimeType] = None


# Schemas used to store the data types
# TODO(lc 09-01-2022) We should ensure that all pydantic models are
# in camel case, but for now, the schemas are following the format
# they are currently saved in on the OT-2 to avoid a large migration.
class TipLengthModel(BaseModel):
    tipLength: float = Field(..., description="Tip length data found from calibration.")
    lastModified: DatetimeType = Field(
        ..., description="The last time this tip length was calibrated."
    )
    source: types.SourceType = Field(
        default=types.SourceType.factory, description="The source of calibration."
    )
    status: CalibrationStatus = Field(
        default_factory=CalibrationStatus,
        description="The status of the calibration data.",
    )
    # Old data may have a `uri` field, replaced later by `definitionHash`.
    # uri: typing.Union[LabwareUri, Literal[""]] = Field(
    #    ..., description="The tiprack URI associated with the tip length data."
    # )
    definitionHash: str = Field(
        ..., description="The tiprack hash associated with the tip length data."
    )

    @field_validator("tipLength")
    @classmethod
    def ensure_tip_length_positive(cls, tipLength: float) -> float:
        if tipLength < 0.0:
            raise ValueError("Tip Length must be a positive number")
        return tipLength


class DeckCalibrationModel(BaseModel):
    attitude: types.AttitudeMatrix = Field(
        ..., description="Attitude matrix for deck found from calibration."
    )
    last_modified: typing.Optional[DatetimeType] = Field(
        default=None, description="The last time this deck was calibrated."
    )
    source: types.SourceType = Field(
        default=types.SourceType.factory, description="The source of calibration."
    )
    pipette_calibrated_with: typing.Optional[str] = Field(
        default=None, description="The pipette id used to calibrate the deck."
    )
    tiprack: typing.Optional[str] = Field(
        default=None, description="The tiprack id used to calibrate the deck."
    )
    status: CalibrationStatus = Field(
        default_factory=CalibrationStatus,
        description="The status of the calibration data.",
    )


class InstrumentOffsetModel(BaseModel):
    offset: Point = Field(..., description="Instrument offset found from calibration.")
    tiprack: str = Field(..., description="Tiprack used to calibrate this offset")
    uri: str = Field(
        ..., description="The URI of the labware used for instrument offset"
    )
    last_modified: DatetimeType = Field(
        ..., description="The last time this instrument was calibrated."
    )
    source: types.SourceType = Field(
        default=types.SourceType.factory, description="The source of calibration."
    )
    status: CalibrationStatus = Field(
        default_factory=CalibrationStatus,
        description="The status of the calibration data.",
    )


# TODO(lc 09-19-2022) We need to refactor the calibration endpoints
# so that we only need to use one data model schema. This model is a
# temporary workaround to match with current behavior.
class PipetteOffsetCalibration(BaseModel):
    pipette: str = Field(..., description="Pipette id associated with calibration.")
    mount: str = Field(
        ..., description="The mount that this pipette was calibrated with."
    )
    offset: Point = Field(..., description="Instrument offset found from calibration.")
    tiprack: str = Field(..., description="Tiprack used to calibrate this offset")
    uri: str = Field(
        ..., description="The URI of the labware used for instrument offset"
    )
    last_modified: DatetimeType = Field(
        ..., description="The last time this instrument was calibrated."
    )
    source: types.SourceType = Field(
        default=types.SourceType.factory, description="The source of calibration."
    )
    status: CalibrationStatus = Field(
        default_factory=CalibrationStatus,
        description="The status of the calibration data.",
    )


# TODO(lc 09-19-2022) We need to refactor the calibration endpoints
# so that we only need to use one data model schema. This model is a
# temporary workaround to match with current behavior.
class TipLengthCalibration(BaseModel):
    pipette: str = Field(..., description="Pipette id associated with calibration.")
    tiprack: str = Field(
        ..., description="The tiprack hash associated with this tip length data."
    )
    tipLength: float = Field(..., description="Tip length data found from calibration.")
    lastModified: DatetimeType = Field(
        ..., description="The last time this tip length was calibrated."
    )
    source: types.SourceType = Field(
        default=types.SourceType.factory, description="The source of calibration."
    )
    status: CalibrationStatus = Field(
        default_factory=CalibrationStatus,
        description="The status of the calibration data.",
    )
    uri: typing.Union[LabwareUri, Literal[""]] = Field(
        ..., description="The tiprack URI associated with the tip length data."
    )

    @field_validator("tipLength")
    @classmethod
    def ensure_tip_length_positive(cls, tipLength: float) -> float:
        if tipLength < 0.0:
            raise ValueError("Tip Length must be a positive number")
        return tipLength



"""Protocol Engine types for hardware passthrough."""

from enum import Enum


class MovementAxis(str, Enum):
    """Axis on which to issue a relative movement."""

    X = "x"
    Y = "y"
    Z = "z"


class MotorAxis(str, Enum):
    """Motor axis on which to issue a home command."""

    X = "x"
    Y = "y"
    LEFT_Z = "leftZ"
    RIGHT_Z = "rightZ"
    LEFT_PLUNGER = "leftPlunger"
    RIGHT_PLUNGER = "rightPlunger"
    EXTENSION_Z = "extensionZ"
    EXTENSION_JAW = "extensionJaw"
    AXIS_96_CHANNEL_CAM = "axis96ChannelCam"



from opentrons.protocol_engine.clients.sync_client import SyncClient
from opentrons.protocol_engine.types import (
    DeckSlotLocation,
    LabwareLocation,
    ModuleLocation,
    OnLabwareLocation,
    AddressableAreaLocation,
    InStackerHopperLocation,
)


def well(engine_client: SyncClient, well_name: str, labware_id: str) -> str:
    """Return a human-readable string representing a well and its location.

    For example: "A1 of My Cool Labware on C2".
    """
    labware_location = OnLabwareLocation(labwareId=labware_id)
    return f"{well_name} of {_labware_location_string(engine_client, labware_location)}"


def _module_in_location_string(module_id: str, engine_client: SyncClient) -> str:
    module_name = engine_client.state.modules.get_definition(
        module_id=module_id
    ).displayName
    module_on = engine_client.state.modules.get_location(module_id=module_id)
    module_on_string = _labware_location_string(engine_client, module_on)
    return f"{module_name} on {module_on_string}"


def _labware_location_string(
    engine_client: SyncClient, location: LabwareLocation
) -> str:
    if isinstance(location, DeckSlotLocation):
        # TODO(mm, 2023-10-11):
        # Ideally, we might want to use the display name specified by the deck definition?
        return f"slot {location.slotName.id}"

    elif isinstance(location, ModuleLocation):
        return _module_in_location_string(location.moduleId, engine_client)

    elif isinstance(location, OnLabwareLocation):
        labware_name = _labware_name(engine_client, location.labwareId)
        labware_on = engine_client.state.labware.get_location(
            labware_id=location.labwareId
        )
        labware_on_string = _labware_location_string(engine_client, labware_on)
        return f"{labware_name} on {labware_on_string}"

    elif isinstance(location, AddressableAreaLocation):
        # In practice this will always be a deck slot or staging slot
        return f"slot {location.addressableAreaName}"

    elif location == "offDeck":
        return "[off-deck]"

    elif location == "systemLocation":
        return "[systemLocation]"

    elif isinstance(location, InStackerHopperLocation):
        return (
            f"stored in {_module_in_location_string(location.moduleId, engine_client)}"
        )


def _labware_name(engine_client: SyncClient, labware_id: str) -> str:
    """Return the user-specified labware label, or fall back to the display name from the def."""
    user_name = engine_client.state.labware.get_user_specified_display_name(
        labware_id=labware_id
    )
    definition_name = engine_client.state.labware.get_definition(
        labware_id=labware_id
    ).metadata.displayName

    return user_name if user_name is not None else definition_name



"""Command models for moving any robot axis to an absolute position."""
from __future__ import annotations
from typing import Literal, Optional, Type, TYPE_CHECKING, Any

from pydantic import Field, BaseModel
from pydantic.json_schema import SkipJsonSchema

from opentrons.hardware_control import HardwareControlAPI
from opentrons.protocol_engine.resources import ensure_ot3_hardware

from .common import MotorAxisMapType, DestinationRobotPositionResult
from ..command import (
    AbstractCommandImpl,
    BaseCommand,
    BaseCommandCreate,
    SuccessData,
)
from ...errors.error_occurrence import ErrorOccurrence

if TYPE_CHECKING:
    from opentrons.protocol_engine.execution import GantryMover


MoveAxesToCommandType = Literal["robot/moveAxesTo"]


def _remove_default(s: dict[str, Any]) -> None:
    s.pop("default", None)


class MoveAxesToParams(BaseModel):
    """Payload required to move axes to absolute position."""

    axis_map: MotorAxisMapType = Field(
        ..., description="The specified axes to move to an absolute deck position with."
    )
    critical_point: MotorAxisMapType | SkipJsonSchema[None] = Field(
        default=None,
        description="The critical point to move the mount with.",
        json_schema_extra=_remove_default,
    )
    speed: float | SkipJsonSchema[None] = Field(
        default=None,
        description="The max velocity to move the axes at. Will fall to hardware defaults if none provided.",
        json_schema_extra=_remove_default,
    )


class MoveAxesToResult(DestinationRobotPositionResult):
    """Result data from the execution of a MoveAxesTo command."""

    pass


class MoveAxesToImplementation(
    AbstractCommandImpl[MoveAxesToParams, SuccessData[MoveAxesToResult]]
):
    """MoveAxesTo command implementation."""

    def __init__(
        self,
        gantry_mover: GantryMover,
        hardware_api: HardwareControlAPI,
        **kwargs: object,
    ) -> None:
        self._gantry_mover = gantry_mover
        self._hardware_api = hardware_api

    async def execute(self, params: MoveAxesToParams) -> SuccessData[MoveAxesToResult]:
        """Move the axes on a flex an absolute distance."""
        # TODO (lc 08-16-2024) implement `move_axes` for OT 2 hardware controller
        # and then we can remove this validation.
        ensure_ot3_hardware(self._hardware_api)
        current_position = await self._gantry_mover.move_axes(
            axis_map=params.axis_map,
            speed=params.speed,
            critical_point=params.critical_point,
        )
        return SuccessData(
            public=MoveAxesToResult(position=current_position),
        )


class MoveAxesTo(BaseCommand[MoveAxesToParams, MoveAxesToResult, ErrorOccurrence]):
    """MoveAxesTo command model."""

    commandType: MoveAxesToCommandType = "robot/moveAxesTo"
    params: MoveAxesToParams
    result: Optional[MoveAxesToResult] = None

    _ImplementationCls: Type[MoveAxesToImplementation] = MoveAxesToImplementation


class MoveAxesToCreate(BaseCommandCreate[MoveAxesToParams]):
    """MoveAxesTo command request model."""

    commandType: MoveAxesToCommandType = "robot/moveAxesTo"
    params: MoveAxesToParams

    _CommandCls: Type[MoveAxesTo] = MoveAxesTo



"""Basic pipette data state and store."""

from __future__ import annotations

import dataclasses
from logging import getLogger
from typing import (
    Dict,
    List,
    Mapping,
    Optional,
    Tuple,
    cast,
)

from typing_extensions import assert_never

from opentrons_shared_data.pipette import pipette_definition
from opentrons_shared_data.pipette.ul_per_mm import calculate_ul_per_mm
from opentrons_shared_data.pipette.types import UlPerMmAction

from opentrons.config.defaults_ot2 import Z_RETRACT_DISTANCE
from opentrons.hardware_control.dev_types import PipetteDict
from opentrons.hardware_control import CriticalPoint
from opentrons.hardware_control.nozzle_manager import (
    NozzleMap,
)
from opentrons.types import MountType, Mount as HwMount, Point, NozzleConfigurationType

from . import update_types, fluid_stack
from .. import errors
from ..types import (
    LoadedPipette,
    MotorAxis,
    FlowRates,
    DeckPoint,
    CurrentWell,
    CurrentAddressableArea,
    CurrentPipetteLocation,
    TipGeometry,
)
from ..actions import (
    Action,
    SetPipetteMovementSpeedAction,
    get_state_updates,
)
from ._abstract_store import HasState, HandlesActions

LOG = getLogger(__name__)


@dataclasses.dataclass(frozen=True)
class HardwarePipette:
    """Hardware pipette data."""

    mount: HwMount
    config: PipetteDict


@dataclasses.dataclass(frozen=True)
class CurrentDeckPoint:
    """The latest deck point and mount the robot has accessed."""

    mount: Optional[MountType]
    deck_point: Optional[DeckPoint]


@dataclasses.dataclass(frozen=True)
class BoundingNozzlesOffsets:
    """Offsets of the bounding nozzles of the pipette."""

    back_left_offset: Point
    front_right_offset: Point


@dataclasses.dataclass(frozen=True)
class PipetteBoundingBoxOffsets:
    """Offsets of the corners of the pipette's bounding box."""

    back_left_corner: Point
    front_right_corner: Point
    back_right_corner: Point
    front_left_corner: Point


@dataclasses.dataclass(frozen=True)
class StaticPipetteConfig:
    """Static config for a pipette."""

    model: str
    serial_number: str
    display_name: str
    min_volume: float
    max_volume: float
    channels: int
    tip_configuration_lookup_table: Dict[
        float, pipette_definition.SupportedTipsDefinition
    ]
    nominal_tip_overlap: Dict[str, float]
    home_position: float
    nozzle_offset_z: float
    pipette_bounding_box_offsets: PipetteBoundingBoxOffsets
    bounding_nozzle_offsets: BoundingNozzlesOffsets
    default_nozzle_map: NozzleMap  # todo(mm, 2024-10-14): unused, remove?
    lld_settings: Optional[Dict[str, Dict[str, float]]]
    plunger_positions: Dict[str, float]
    shaft_ul_per_mm: float
    available_sensors: pipette_definition.AvailableSensorDefinition


@dataclasses.dataclass
class PipetteState:
    """Basic pipette data state and getter methods."""

    # todo(mm, 2024-10-14): It's getting difficult to ensure that all of these
    # attributes are populated at the appropriate times. Refactor to a
    # single dict-of-many-things instead of many dicts-of-single-things.
    pipettes_by_id: Dict[str, LoadedPipette]
    pipette_contents_by_id: Dict[str, Optional[fluid_stack.FluidStack]]
    current_location: Optional[CurrentPipetteLocation]
    current_deck_point: CurrentDeckPoint
    attached_tip_by_id: Dict[str, Optional[TipGeometry]]
    movement_speed_by_id: Dict[str, Optional[float]]
    static_config_by_id: Dict[str, StaticPipetteConfig]
    flow_rates_by_id: Dict[str, FlowRates]
    nozzle_configuration_by_id: Dict[str, NozzleMap]
    liquid_presence_detection_by_id: Dict[str, bool]
    ready_to_aspirate_by_id: Dict[str, bool]
    has_clean_tips_by_id: Dict[str, bool]


class PipetteStore(HasState[PipetteState], HandlesActions):
    """Pipette state container."""

    _state: PipetteState

    def __init__(self) -> None:
        """Initialize a PipetteStore and its state."""
        self._state = PipetteState(
            pipettes_by_id={},
            pipette_contents_by_id={},
            attached_tip_by_id={},
            current_location=None,
            current_deck_point=CurrentDeckPoint(mount=None, deck_point=None),
            movement_speed_by_id={},
            static_config_by_id={},
            flow_rates_by_id={},
            nozzle_configuration_by_id={},
            liquid_presence_detection_by_id={},
            ready_to_aspirate_by_id={},
            has_clean_tips_by_id={},
        )

    def handle_action(self, action: Action) -> None:
        """Modify state in reaction to an action."""
        for state_update in get_state_updates(action):
            self._set_load_pipette(state_update)
            self._update_current_location(state_update)
            self._update_pipette_config(state_update)
            self._update_pipette_nozzle_map(state_update)
            self._update_tip_state(state_update)
            self._update_volumes(state_update)
            self._update_ready_for_aspirate(state_update)

        if isinstance(action, SetPipetteMovementSpeedAction):
            self._state.movement_speed_by_id[action.pipette_id] = action.speed

    def _set_load_pipette(self, state_update: update_types.StateUpdate) -> None:
        if state_update.loaded_pipette != update_types.NO_CHANGE:
            pipette_id = state_update.loaded_pipette.pipette_id

            self._state.pipettes_by_id[pipette_id] = LoadedPipette(
                id=pipette_id,
                pipetteName=state_update.loaded_pipette.pipette_name,
                mount=state_update.loaded_pipette.mount,
            )
            self._state.liquid_presence_detection_by_id[pipette_id] = (
                state_update.loaded_pipette.liquid_presence_detection or False
            )
            self._state.movement_speed_by_id[pipette_id] = None
            self._state.attached_tip_by_id[pipette_id] = None
            self._state.ready_to_aspirate_by_id[pipette_id] = False

    def _update_tip_state(self, state_update: update_types.StateUpdate) -> None:
        if state_update.pipette_tip_state != update_types.NO_CHANGE:
            pipette_id = state_update.pipette_tip_state.pipette_id
            if state_update.pipette_tip_state.tip_geometry:
                attached_tip = state_update.pipette_tip_state.tip_geometry

                self._state.attached_tip_by_id[pipette_id] = attached_tip

                static_config = self._state.static_config_by_id.get(pipette_id)
                if static_config:
                    try:
                        tip_configuration = (
                            static_config.tip_configuration_lookup_table[
                                attached_tip.volume
                            ]
                        )
                    except KeyError:
                        # TODO(seth,9/11/2023): this is a bad way of doing defaults but better than max volume.
                        # we used to look up a default tip config via the pipette max volume, but if that isn't
                        # tip volume (as it isn't when we're in low-volume mode) then that lookup fails. Using
                        # the first entry in the table is ok I guess but we really need to generally rethink how
                        # we identify tip classes - looking things up by volume is not enough.
                        tip_configuration = list(
                            static_config.tip_configuration_lookup_table.values()
                        )[0]
                    self._state.flow_rates_by_id[pipette_id] = FlowRates(
                        default_blow_out=tip_configuration.default_blowout_flowrate.values_by_api_level,
                        default_aspirate=tip_configuration.default_aspirate_flowrate.values_by_api_level,
                        default_dispense=tip_configuration.default_dispense_flowrate.values_by_api_level,
                    )

            else:
                pipette_id = state_update.pipette_tip_state.pipette_id
                self._state.attached_tip_by_id[pipette_id] = None
                self._state.has_clean_tips_by_id[pipette_id] = False

                static_config = self._state.static_config_by_id.get(pipette_id)
                if static_config:
                    # TODO(seth,9/11/2023): bad way to do defaulting, see above.
                    tip_configuration = list(
                        static_config.tip_configuration_lookup_table.values()
                    )[0]
                    self._state.flow_rates_by_id[pipette_id] = FlowRates(
                        default_blow_out=tip_configuration.default_blowout_flowrate.values_by_api_level,
                        default_aspirate=tip_configuration.default_aspirate_flowrate.values_by_api_level,
                        default_dispense=tip_configuration.default_dispense_flowrate.values_by_api_level,
                    )

    def _update_current_location(self, state_update: update_types.StateUpdate) -> None:
        location_update = state_update.pipette_location

        if location_update is update_types.NO_CHANGE:
            pass
        elif location_update is update_types.CLEAR:
            self._state.current_location = None
            self._state.current_deck_point = CurrentDeckPoint(
                mount=None, deck_point=None
            )
        else:
            new_logical_location = location_update.new_location
            new_deck_point = location_update.new_deck_point
            match new_logical_location:
                case update_types.Well(labware_id=labware_id, well_name=well_name):
                    self._state.current_location = CurrentWell(
                        pipette_id=location_update.pipette_id,
                        labware_id=labware_id,
                        well_name=well_name,
                    )
                case update_types.AddressableArea(
                    addressable_area_name=addressable_area_name
                ):
                    self._state.current_location = CurrentAddressableArea(
                        pipette_id=location_update.pipette_id,
                        addressable_area_name=addressable_area_name,
                    )
                case None:
                    self._state.current_location = None
                case update_types.NO_CHANGE:
                    pass
            if new_deck_point is not update_types.NO_CHANGE:
                loaded_pipette = self._state.pipettes_by_id[location_update.pipette_id]
                self._state.current_deck_point = CurrentDeckPoint(
                    mount=loaded_pipette.mount, deck_point=new_deck_point
                )

    def _update_pipette_config(self, state_update: update_types.StateUpdate) -> None:
        if state_update.pipette_config != update_types.NO_CHANGE:
            config = state_update.pipette_config.config
            self._state.static_config_by_id[
                state_update.pipette_config.pipette_id
            ] = StaticPipetteConfig(
                serial_number=state_update.pipette_config.serial_number,
                model=config.model,
                display_name=config.display_name,
                min_volume=config.min_volume,
                max_volume=config.max_volume,
                channels=config.channels,
                tip_configuration_lookup_table=config.tip_configuration_lookup_table,
                nominal_tip_overlap=config.nominal_tip_overlap,
                home_position=config.home_position,
                nozzle_offset_z=config.nozzle_offset_z,
                pipette_bounding_box_offsets=PipetteBoundingBoxOffsets(
                    back_left_corner=config.back_left_corner_offset,
                    front_right_corner=config.front_right_corner_offset,
                    back_right_corner=Point(
                        config.front_right_corner_offset.x,
                        config.back_left_corner_offset.y,
                        config.back_left_corner_offset.z,
                    ),
                    front_left_corner=Point(
                        config.back_left_corner_offset.x,
                        config.front_right_corner_offset.y,
                        config.back_left_corner_offset.z,
                    ),
                ),
                bounding_nozzle_offsets=BoundingNozzlesOffsets(
                    back_left_offset=config.nozzle_map.back_left_nozzle_offset,
                    front_right_offset=config.nozzle_map.front_right_nozzle_offset,
                ),
                default_nozzle_map=config.nozzle_map,
                lld_settings=config.pipette_lld_settings,
                plunger_positions=config.plunger_positions,
                shaft_ul_per_mm=config.shaft_ul_per_mm,
                available_sensors=config.available_sensors,
            )
            self._state.flow_rates_by_id[
                state_update.pipette_config.pipette_id
            ] = config.flow_rates
            self._state.nozzle_configuration_by_id[
                state_update.pipette_config.pipette_id
            ] = config.nozzle_map

    def _update_pipette_nozzle_map(
        self, state_update: update_types.StateUpdate
    ) -> None:
        if state_update.pipette_nozzle_map != update_types.NO_CHANGE:
            self._state.nozzle_configuration_by_id[
                state_update.pipette_nozzle_map.pipette_id
            ] = state_update.pipette_nozzle_map.nozzle_map

    def _update_ready_for_aspirate(
        self, state_update: update_types.StateUpdate
    ) -> None:
        if state_update.ready_to_aspirate != update_types.NO_CHANGE:
            self._state.ready_to_aspirate_by_id[
                state_update.ready_to_aspirate.pipette_id
            ] = state_update.ready_to_aspirate.ready_to_aspirate

    def _update_volumes(self, state_update: update_types.StateUpdate) -> None:
        if state_update.pipette_aspirated_fluid == update_types.NO_CHANGE:
            return
        # set the tip state to unclean, if an "empty" update has a clean_tip flag
        # it will set it to true
        self._state.has_clean_tips_by_id[
            state_update.pipette_aspirated_fluid.pipette_id
        ] = False

        if state_update.pipette_aspirated_fluid.type == "aspirated":
            self._update_aspirated(state_update.pipette_aspirated_fluid)
        elif state_update.pipette_aspirated_fluid.type == "ejected":
            self._update_ejected(state_update.pipette_aspirated_fluid)
        elif state_update.pipette_aspirated_fluid.type == "empty":
            self._update_empty(state_update.pipette_aspirated_fluid)
        elif state_update.pipette_aspirated_fluid.type == "unknown":
            self._update_unknown(state_update.pipette_aspirated_fluid)
        else:
            assert_never(state_update.pipette_aspirated_fluid.type)

    def _update_aspirated(
        self, update: update_types.PipetteAspiratedFluidUpdate
    ) -> None:
        if self._state.pipette_contents_by_id[update.pipette_id] is None:
            self._state.pipette_contents_by_id[
                update.pipette_id
            ] = fluid_stack.FluidStack()

        self._fluid_stack_log_if_empty(update.pipette_id).add_fluid(update.fluid)

    def _update_ejected(self, update: update_types.PipetteEjectedFluidUpdate) -> None:
        self._fluid_stack_log_if_empty(update.pipette_id).remove_fluid(update.volume)

    def _update_empty(self, update: update_types.PipetteEmptyFluidUpdate) -> None:
        self._state.pipette_contents_by_id[update.pipette_id] = fluid_stack.FluidStack()
        self._state.has_clean_tips_by_id[update.pipette_id] = update.clean_tip

    def _update_unknown(self, update: update_types.PipetteUnknownFluidUpdate) -> None:
        self._state.pipette_contents_by_id[update.pipette_id] = None

    def _fluid_stack_log_if_empty(self, pipette_id: str) -> fluid_stack.FluidStack:
        stack = self._state.pipette_contents_by_id[pipette_id]
        if stack is None:
            LOG.error("Pipette state tried to alter an unknown-contents pipette")
            return fluid_stack.FluidStack()
        return stack


class PipetteView:
    """Read-only view of computed pipettes state."""

    _state: PipetteState

    def __init__(self, state: PipetteState) -> None:
        """Initialize the view with its backing state value."""
        self._state = state

    def get(self, pipette_id: str) -> LoadedPipette:
        """Get pipette data by the pipette's unique identifier."""
        try:
            return self._state.pipettes_by_id[pipette_id]
        except KeyError as e:
            raise errors.PipetteNotLoadedError(
                f"Pipette {pipette_id} not found."
            ) from e

    def get_mount(self, pipette_id: str) -> MountType:
        """Get the pipette's mount."""
        return self.get(pipette_id).mount

    def get_all(self) -> List[LoadedPipette]:
        """Get a list of all pipette entries in state."""
        return list(self._state.pipettes_by_id.values())

    def get_by_mount(self, mount: MountType) -> Optional[LoadedPipette]:
        """Get pipette data by the pipette's mount."""
        for pipette in self._state.pipettes_by_id.values():
            if pipette.mount == mount:
                return pipette
        return None

    def get_hardware_pipette(
        self,
        pipette_id: str,
        attached_pipettes: Mapping[HwMount, Optional[PipetteDict]],
    ) -> HardwarePipette:
        """Get a pipette's hardware configuration and state by ID."""
        pipette_data = self.get(pipette_id)
        pipette_name = pipette_data.pipetteName
        mount = pipette_data.mount

        hw_mount = mount.to_hw_mount()
        hw_config = attached_pipettes[hw_mount]

        # TODO(mc, 2022-01-11): HW controller may return an empty dict for
        # no pipette attached instead of `None`. Update when fixed in HWAPI
        if not hw_config:
            raise errors.PipetteNotAttachedError(f"No pipette attached on {mount}")

        elif (
            hw_config["name"] != pipette_name
            and pipette_name not in hw_config["back_compat_names"]
        ):
            raise errors.PipetteNotAttachedError(
                f"Found {hw_config['name']} on {mount}, "
                f"but {pipette_id} is a {pipette_name}"
            )

        return HardwarePipette(mount=hw_mount, config=hw_config)

    def get_current_location(self) -> Optional[CurrentPipetteLocation]:
        """Get the last accessed location and which pipette accessed it."""
        return self._state.current_location

    def get_deck_point(self, pipette_id: str) -> Optional[DeckPoint]:
        """Get the deck point of a pipette by ID, or None if it was not associated with the last move operation."""
        loaded_pipette = self.get(pipette_id)
        current_deck_point = self._state.current_deck_point
        if loaded_pipette.mount == current_deck_point.mount:
            return current_deck_point.deck_point
        return None

    def get_attached_tip(self, pipette_id: str) -> Optional[TipGeometry]:
        """Get details of the pipette's attached tip.

        Returns:
            The tip's volume and length, or None if there is no tip attached,
        """
        try:
            return self._state.attached_tip_by_id[pipette_id]
        except KeyError as e:
            raise errors.PipetteNotLoadedError(
                f"Pipette {pipette_id} no found; unable to get attached tip."
            ) from e

    def get_all_attached_tips(self) -> List[Tuple[str, TipGeometry]]:
        """Get a list of all attached tips.

        Returns:
            A list of pipette ID, tip details tuples.
        """
        return [
            (pipette_id, tip)
            for pipette_id, tip in self._state.attached_tip_by_id.items()
            if tip is not None
        ]

    def get_aspirated_volume(self, pipette_id: str) -> Optional[float]:
        """Get the currently aspirated volume of a pipette by ID.

        This is the volume currently displaced by the plunger relative to its bottom position,
        regardless of whether that volume likely contains liquid or air. This makes it the right
        function to call to know how much more volume the plunger may displace.

        Returns:
            The volume the pipette has aspirated.
            None, after blow-out and the plunger is in an unsafe position.

        Raises:
            PipetteNotLoadedError: pipette ID does not exist.
            TipNotAttachedError: if no tip is attached to the pipette.
        """
        self.validate_tip_state(pipette_id, True)

        try:
            stack = self._state.pipette_contents_by_id[pipette_id]
            if stack is None:
                return None
            return stack.aspirated_volume()

        except KeyError as e:
            raise errors.PipetteNotLoadedError(
                f"Pipette {pipette_id} not found; unable to get current volume."
            ) from e

    def get_has_clean_tip(self, pipette_id: str) -> bool:
        """Get if the tip of a pipette by ID is clean.

        This is only true directly after a pick up tip, once any kind of aspirate happens
        it is no longer clean

        Returns:
            True if the tip is clean
            False if it is unclean

        Raises:
            PipetteNotLoadedError: pipette ID does not exist.
            TipNotAttachedError: if no tip is attached to the pipette.
        """
        self.validate_tip_state(pipette_id, True)

        try:
            return self._state.has_clean_tips_by_id[pipette_id]
        except KeyError as e:
            raise errors.PipetteNotLoadedError(
                f"Pipette {pipette_id} not found; unable to get current volume."
            ) from e

    def get_liquid_dispensed_by_ejecting_volume(
        self, pipette_id: str, volume: float
    ) -> Optional[float]:
        """Get the amount of liquid (not air) that will be dispensed if the pipette ejects a specified volume.

        For instance, if the pipette contains, in vertical order,
        10 ul air
        80 ul liquid
        5 ul air

        then dispensing 10ul would result in 5ul of liquid; dispensing 85 ul would result in 80ul liquid; dispensing
        95ul would result in 80ul liquid.

        Returns:
            The volume of liquid that would be dispensed by the requested volume.
            None, after blow-out or when the plunger is in an unsafe position.

        Raises:
            PipetteNotLoadedError: pipette ID does not exist.
            TipnotAttachedError: No tip is attached to the pipette.
        """
        self.validate_tip_state(pipette_id, True)

        try:
            stack = self._state.pipette_contents_by_id[pipette_id]
            if stack is None:
                return None
            return stack.liquid_part_of_dispense_volume(volume)

        except KeyError as e:
            raise errors.PipetteNotLoadedError(
                f"Pipette {pipette_id} not found; unable to get current liquid volume."
            ) from e

    def get_working_volume(self, pipette_id: str) -> float:
        """Get the working maximum volume of a pipette by ID.

        Raises:
            PipetteNotLoadedError: pipette ID does not exist.
            TipNotAttachedError: if no tip is attached to the pipette.
        """
        max_volume = self.get_maximum_volume(pipette_id)
        attached_tip = self.get_attached_tip(pipette_id)

        if not attached_tip:
            raise errors.TipNotAttachedError(
                f"Pipette {pipette_id} has no tip attached; unable to calculate working maximum volume."
            )

        return min(attached_tip.volume, max_volume)

    def get_available_volume(self, pipette_id: str) -> Optional[float]:
        """Get the available volume of a pipette by ID."""
        working_volume = self.get_working_volume(pipette_id)
        current_volume = self.get_aspirated_volume(pipette_id)

        return max(0.0, working_volume - current_volume) if current_volume else None

    def get_pipette_lld_settings(
        self, pipette_id: str
    ) -> Optional[Dict[str, Dict[str, float]]]:
        """Get the liquid level settings for all possible tips for a single pipette."""
        return self.get_config(pipette_id).lld_settings

    def get_current_tip_lld_settings(self, pipette_id: str) -> float:
        """Get the liquid level settings for pipette and its current tip."""
        attached_tip = self.get_attached_tip(pipette_id)
        if attached_tip is None or attached_tip.volume is None:
            return 0
        lld_settings = self.get_pipette_lld_settings(pipette_id)
        tipVolume = "t" + str(int(attached_tip.volume))
        if (
            lld_settings is None
            or lld_settings[tipVolume] is None
            or lld_settings[tipVolume]["minHeight"] is None
        ):
            return 0
        return float(lld_settings[tipVolume]["minHeight"])

    def validate_tip_state(self, pipette_id: str, expected_has_tip: bool) -> None:
        """Validate that a pipette's tip state matches expectations."""
        attached_tip = self.get_attached_tip(pipette_id)

        if expected_has_tip is True and attached_tip is None:
            raise errors.TipNotAttachedError(
                "Pipette should have a tip attached, but does not."
            )
        if expected_has_tip is False and attached_tip is not None:
            raise errors.TipAttachedError(
                "Pipette should not have a tip attached, but does."
            )

    def get_movement_speed(
        self, pipette_id: str, requested_speed: Optional[float] = None
    ) -> Optional[float]:
        """Return the given pipette's requested or current movement speed."""
        return requested_speed or self._state.movement_speed_by_id[pipette_id]

    def get_config(self, pipette_id: str) -> StaticPipetteConfig:
        """Get the static pipette configuration by pipette id."""
        try:
            return self._state.static_config_by_id[pipette_id]
        except KeyError as e:
            raise errors.PipetteNotLoadedError(
                f"Pipette {pipette_id} not found; unable to get pipette configuration."
            ) from e

    def get_model_name(self, pipette_id: str) -> str:
        """Return the given pipette's model name."""
        return self.get_config(pipette_id).model

    def get_display_name(self, pipette_id: str) -> str:
        """Return the given pipette's display name."""
        return self.get_config(pipette_id).display_name

    def get_serial_number(self, pipette_id: str) -> str:
        """Get the serial number of the pipette."""
        return self.get_config(pipette_id).serial_number

    def get_channels(self, pipette_id: str) -> int:
        """Return the max channels of the pipette."""
        return self.get_config(pipette_id).channels

    def get_active_channels(self, pipette_id: str) -> int:
        """Get the number of channels being used in the given pipette's configuration."""
        return self.get_nozzle_configuration(pipette_id).tip_count

    def get_minimum_volume(self, pipette_id: str) -> float:
        """Return the given pipette's minimum volume."""
        return self.get_config(pipette_id).min_volume

    def get_maximum_volume(self, pipette_id: str) -> float:
        """Return the given pipette's maximum volume."""
        return self.get_config(pipette_id).max_volume

    def get_instrument_max_height_ot2(self, pipette_id: str) -> float:
        """Get calculated max instrument height for an OT-2."""
        config = self.get_config(pipette_id)
        return config.home_position - Z_RETRACT_DISTANCE + config.nozzle_offset_z

    def get_return_tip_scale(self, pipette_id: str) -> float:
        """Return the given pipette's return tip height scale."""
        max_volume = self.get_maximum_volume(pipette_id)
        working_volume = max_volume
        tip = self.get_attached_tip(pipette_id)
        if tip:
            working_volume = tip.volume

        if working_volume in self.get_config(pipette_id).tip_configuration_lookup_table:
            tip_lookup = self.get_config(pipette_id).tip_configuration_lookup_table[
                working_volume
            ]
        else:
            tip_lookup = self.get_config(pipette_id).tip_configuration_lookup_table[
                max_volume
            ]
        return tip_lookup.default_return_tip_height

    def get_flow_rates(self, pipette_id: str) -> FlowRates:
        """Get the default flow rates for the pipette."""
        try:
            return self._state.flow_rates_by_id[pipette_id]
        except KeyError as e:
            raise errors.PipetteNotLoadedError(
                f"Pipette {pipette_id} not found; unable to get pipette flow rates."
            ) from e

    def get_nominal_tip_overlap(self, pipette_id: str, labware_uri: str) -> float:
        """Get the nominal tip overlap for a given labware from config."""
        tip_overlaps_by_uri = self.get_config(pipette_id).nominal_tip_overlap

        try:
            return tip_overlaps_by_uri[labware_uri]
        except KeyError:
            return tip_overlaps_by_uri.get("default", 0)

    def get_z_axis(self, pipette_id: str) -> MotorAxis:
        """Get the MotorAxis representing this pipette's Z stage."""
        mount = self.get(pipette_id).mount
        return MotorAxis.LEFT_Z if mount == MountType.LEFT else MotorAxis.RIGHT_Z

    def get_plunger_axis(self, pipette_id: str) -> MotorAxis:
        """Get the MotorAxis representing this pipette's plunger."""
        mount = self.get(pipette_id).mount
        return (
            MotorAxis.LEFT_PLUNGER
            if mount == MountType.LEFT
            else MotorAxis.RIGHT_PLUNGER
        )

    def get_nozzle_layout_type(self, pipette_id: str) -> NozzleConfigurationType:
        """Get the current set nozzle layout configuration."""
        nozzle_map_for_pipette = self._state.nozzle_configuration_by_id[pipette_id]
        return nozzle_map_for_pipette.configuration

    def get_is_partially_configured(self, pipette_id: str) -> bool:
        """Determine if the provided pipette is partially configured."""
        return self.get_nozzle_layout_type(pipette_id) != NozzleConfigurationType.FULL

    def get_primary_nozzle(self, pipette_id: str) -> str:
        """Get the primary nozzle, if any, related to the given pipette's nozzle configuration."""
        nozzle_map = self._state.nozzle_configuration_by_id[pipette_id]
        return nozzle_map.starting_nozzle

    def get_nozzle_configurations(self) -> Dict[str, NozzleMap]:
        """Get the nozzle maps of all pipettes, keyed by pipette ID."""
        return self._state.nozzle_configuration_by_id.copy()

    def get_nozzle_configuration(self, pipette_id: str) -> NozzleMap:
        """Get the nozzle map of the pipette."""
        return self._state.nozzle_configuration_by_id[pipette_id]

    def _get_critical_point_offset_without_tip(
        self, pipette_id: str, critical_point: Optional[CriticalPoint]
    ) -> Point:
        """Get the offset of the specified critical point from pipette's mount position."""
        nozzle_map = self._state.nozzle_configuration_by_id[pipette_id]
        match critical_point:
            case CriticalPoint.INSTRUMENT_XY_CENTER:
                return nozzle_map.instrument_xy_center_offset
            case CriticalPoint.XY_CENTER:
                return nozzle_map.xy_center_offset
            case CriticalPoint.Y_CENTER:
                return nozzle_map.y_center_offset
            case CriticalPoint.FRONT_NOZZLE:
                return nozzle_map.front_nozzle_offset
            case _:
                return nozzle_map.starting_nozzle_offset

    def get_pipette_bounding_nozzle_offsets(
        self, pipette_id: str
    ) -> BoundingNozzlesOffsets:
        """Get the nozzle offsets of the pipette's bounding nozzles."""
        return self.get_config(pipette_id).bounding_nozzle_offsets

    def get_pipette_bounding_box(self, pipette_id: str) -> PipetteBoundingBoxOffsets:
        """Get the bounding box of the pipette."""
        return self.get_config(pipette_id).pipette_bounding_box_offsets

    # TODO (spp, 2024-09-17): in order to find the position of pipette at destination,
    #  this method repeats the same steps that waypoints builder does while finding
    #  waypoints to move to. We should consolidate these steps into a shared entity
    #  so that the deck conflict checker and movement plan builder always remain in sync.
    def get_pipette_bounds_at_specified_move_to_position(
        self,
        pipette_id: str,
        destination_position: Point,
        critical_point: Optional[CriticalPoint],
    ) -> Tuple[Point, Point, Point, Point]:
        """Get the pipette's bounding box position when critical point is at the destination position.

        Returns a tuple of the pipette's bounding box position in deck coordinates as-
            (back_left_bound, front_right_bound, back_right_bound, front_left_bound)
        Bounding box of the pipette includes the pipette's outer casing as well as nozzles.
        """
        tip = self.get_attached_tip(pipette_id)

        # *Offset* of pipette's critical point w.r.t pipette mount
        critical_point_offset = self._get_critical_point_offset_without_tip(
            pipette_id, critical_point
        )

        # Position of the above critical point at destination, in deck coordinates
        critical_point_position = destination_position + Point(
            x=0, y=0, z=tip.length if tip else 0
        )

        # Get the pipette bounding box coordinates
        pipette_bounds_offsets = self.get_config(
            pipette_id
        ).pipette_bounding_box_offsets
        pip_back_left_bound = (
            critical_point_position
            - critical_point_offset
            + pipette_bounds_offsets.back_left_corner
        )
        pip_front_right_bound = (
            critical_point_position
            - critical_point_offset
            + pipette_bounds_offsets.front_right_corner
        )
        pip_back_right_bound = Point(
            pip_front_right_bound.x, pip_back_left_bound.y, pip_front_right_bound.z
        )
        pip_front_left_bound = Point(
            pip_back_left_bound.x, pip_front_right_bound.y, pip_back_left_bound.z
        )
        return (
            pip_back_left_bound,
            pip_front_right_bound,
            pip_back_right_bound,
            pip_front_left_bound,
        )

    def get_pipette_supports_pressure(self, pipette_id: str) -> bool:
        """Return if this pipette supports a pressure sensor."""
        return (
            "pressure"
            in self._state.static_config_by_id[pipette_id].available_sensors.sensors
        )

    def get_liquid_presence_detection(self, pipette_id: str) -> bool:
        """Determine if liquid presence detection is enabled for this pipette."""
        try:
            return self._state.liquid_presence_detection_by_id[pipette_id]
        except KeyError as e:
            raise errors.PipetteNotLoadedError(
                f"Pipette {pipette_id} not found; unable to determine if pipette liquid presence detection enabled."
            ) from e

    def get_nozzle_configuration_supports_lld(self, pipette_id: str) -> bool:
        """Determine if the current partial tip configuration supports LLD."""
        nozzle_map = self.get_nozzle_configuration(pipette_id)
        if (
            nozzle_map.physical_nozzle_count == 96
            and nozzle_map.back_left != nozzle_map.full_instrument_back_left
            and nozzle_map.front_right != nozzle_map.full_instrument_front_right
        ):
            return False
        return True

    def lookup_volume_to_mm_conversion(
        self, pipette_id: str, volume: float, action: str
    ) -> float:
        """Get the volumn to mm conversion for a pipette."""
        try:
            lookup_volume = self.get_working_volume(pipette_id)
        except errors.TipNotAttachedError:
            lookup_volume = self.get_maximum_volume(pipette_id)

        pipette_config = self.get_config(pipette_id)
        lookup_table_from_config = pipette_config.tip_configuration_lookup_table
        try:
            tip_settings = lookup_table_from_config[lookup_volume]
        except KeyError:
            tip_settings = list(lookup_table_from_config.values())[0]
        return calculate_ul_per_mm(
            volume,
            cast(UlPerMmAction, action),
            tip_settings,
            shaft_ul_per_mm=pipette_config.shaft_ul_per_mm,
        )

    def lookup_plunger_position_name(
        self, pipette_id: str, position_name: str
    ) -> float:
        """Get the plunger position provided for the given pipette id."""
        return self.get_config(pipette_id).plunger_positions[position_name]

    def get_ready_to_aspirate(self, pipette_id: str) -> bool:
        """Get if the provided pipette is ready to aspirate for the given pipette id."""
        try:
            return self._state.ready_to_aspirate_by_id[pipette_id]
        except KeyError as e:
            raise errors.PipetteNotLoadedError(
                f"Pipette {pipette_id} not found; unable to determine if pipette ready to aspirate."
            ) from e



"""Functions and utilites for OT3 calibration."""
from __future__ import annotations
from functools import lru_cache
from dataclasses import dataclass
from typing_extensions import Final, Literal, TYPE_CHECKING
from typing import Tuple, List, Dict, Any, Optional, Union
import datetime
import numpy as np
from enum import Enum
from math import floor, copysign
from logging import getLogger
from opentrons.util.linal import solve_attitude, SolvePoints, DoubleMatrix

from .types import OT3Mount, Axis, GripperProbe
from opentrons.types import Point
from opentrons.config.types import CapacitivePassSettings, EdgeSenseSettings, OT3Config
from opentrons.hardware_control.types import InstrumentProbeType
import json

from opentrons_shared_data.deck import (
    get_calibration_square_position_in_slot,
    Z_PREP_OFFSET,
    CALIBRATION_PROBE_RADIUS,
    CALIBRATION_SQUARE_EDGES as SQUARE_EDGES,
)
from opentrons_shared_data.errors.exceptions import (
    CalibrationStructureNotFoundError,
    EdgeNotFoundError,
    EarlyCapacitiveSenseTrigger,
    InaccurateNonContactSweepError,
    MisalignedGantryError,
)
from .robot_calibration import (
    RobotCalibration,
    DeckCalibration,
)
from opentrons.calibration_storage import types
from opentrons.calibration_storage.ot3.deck_attitude import (
    save_robot_belt_attitude,
    get_robot_belt_attitude,
    delete_robot_belt_attitude,
)
from opentrons.config.robot_configs import (
    default_ot3_deck_calibration,
)
from opentrons.config import defaults_ot3
from .util import DeckTransformState

if TYPE_CHECKING:
    from opentrons.hardware_control import OT3HardwareControlAPI

LOG = getLogger(__name__)

LINEAR_TRANSIT_HEIGHT: Final[float] = 1
SEARCH_TRANSIT_HEIGHT: Final[float] = 5
GRIPPER_GRIP_FORCE: Final[float] = 20  # FIXME: (andy s) this adds error, reduce to 5N

SLOT_CENTER = 5
SLOT_FRONT_LEFT = 1
SLOT_FRONT_RIGHT = 3
SLOT_REAR_LEFT = 10

PREP_OFFSET_DEPTH = Point(*Z_PREP_OFFSET)
EDGES = {
    "left": Point(*SQUARE_EDGES["left"]),
    "right": Point(*SQUARE_EDGES["right"]),
    "top": Point(*SQUARE_EDGES["top"]),
    "bottom": Point(*SQUARE_EDGES["bottom"]),
}
OFFSET_SECONDARY_PROBE = {
    8: Point(x=0, y=9 * 7, z=0),
    96: Point(x=9 * -11, y=9 * 7, z=0),
}


class CalibrationMethod(Enum):
    BINARY_SEARCH = "binary search"
    NONCONTACT_PASS = "noncontact pass"


class CalibrationTarget(Enum):
    DECK_OBJECT = "deck_object"
    GANTRY_INSTRUMENT = "gantry_instrument"


@dataclass
class CalibrationSlot:
    slot: int
    nominal: Point
    actual: Point


class AlignmentShift(Enum):
    LEFT_TO_RIGHT_Y = "left_to_right_y"
    LEFT_TO_RIGHT_Z = "left_to_right_z"
    FRONT_TO_REAR_X = "front_to_rear_x"
    FRONT_TO_REAR_Z = "front_to_rear_z"


# 96ch is 99mm from left->right, and 63mm front->rear
# deck calibration squares are 328mm left->right and 321mm front->rear
# we need to support <=0.1mm shift from 96ch left->right
# which means we would ideally spec left/right shift <=0.33mm
# and front/rear shift <=0.5 (but in reality will be bigger)
# TODO: these will need to update (increase) after testing
#       on DVT2 lifetime units, as well as PVT units
MAX_SHIFT = {
    AlignmentShift.LEFT_TO_RIGHT_Y: 0.5,  # increased from 0.3, based on test results
    AlignmentShift.LEFT_TO_RIGHT_Z: 0.5,  # increased from 0.3, based on test results
    AlignmentShift.FRONT_TO_REAR_X: 0.5,
    AlignmentShift.FRONT_TO_REAR_Z: 0.5,
}


def _verify_height(
    found_pos: float, expected_pos: float, settings: EdgeSenseSettings
) -> None:
    """
    Evaluate the height found by capacitive probe against search settings.
    """
    if found_pos > expected_pos + settings.early_sense_tolerance_mm:
        raise EarlyCapacitiveSenseTrigger(found_pos, expected_pos)


async def _verify_edge_pos(
    hcapi: OT3HardwareControlAPI,
    mount: OT3Mount,
    search_axis: Union[Literal[Axis.X, Axis.Y]],
    found_edge: Point,
    last_stride: float,
    search_direction: Literal[1, -1],
    probe: InstrumentProbeType = InstrumentProbeType.PRIMARY,
) -> None:
    """
    Probe both sides of the found edge in the search axis and compare the results.
    If the position found is valid, we should see that the probe hit the deck on
    one side and miss on the other. If the results are not opposite of each other,
    we didn't find the edge properly.
    """
    edge_settings = hcapi.config.calibration.edge_sense
    # twice the last stride size
    check_stride = last_stride * 2
    last_result = None
    edge_name_str = f"{'+' if search_direction == -1 else '-'}{search_axis}"
    for dir in [-1, 1]:
        checking_pos = found_edge + search_axis.set_in_point(
            Point(0, 0, 0), check_stride * dir
        )
        LOG.info(
            f"Checking {edge_name_str} in {dir} direction at {checking_pos}, stride_size: {check_stride}"
        )
        height, hit_deck = await _probe_deck_at(
            hcapi, mount, checking_pos, edge_settings.pass_settings, probe=probe
        )
        _verify_height(height, found_edge.z, edge_settings)
        LOG.info(f"Deck {'hit' if hit_deck else 'miss'} at check pos: {checking_pos}")
        if last_result is not None and hit_deck != last_result:
            LOG.info(
                f"Edge {edge_name_str} verified successfully at {check_stride} mm resolution."
            )
            return
        else:
            last_result = hit_deck
    raise EdgeNotFoundError(edge_name_str, check_stride)


def critical_edge_offset(
    search_axis: Union[Literal[Axis.X, Axis.Y]], direction_if_hit: Literal[1, -1]
) -> Point:
    """
    Offset to be applied when we are aligning the edge of the probe to the edge of the
    calibration square.
    """
    return search_axis.set_in_point(
        Point(0, 0, 0), direction_if_hit * CALIBRATION_PROBE_RADIUS
    )


async def find_edge_binary(
    hcapi: OT3HardwareControlAPI,
    mount: OT3Mount,
    slot_edge_nominal: Point,
    search_axis: Union[Literal[Axis.X, Axis.Y]],
    direction_if_hit: Literal[1, -1],
    raise_verify_error: bool = True,
    probe: InstrumentProbeType = InstrumentProbeType.PRIMARY,
) -> Point:
    """
    Find the true position of one edge of the calibration slot in the deck.
    The nominal position of the calibration slots is known because they're
    machined into the deck, but if we haven't yet calibrated we won't know
    quite where they are. This routine finds the XY position that will
    place the calibration probe such that its center is in the slot, and
    one edge is on the edge of the slot.
    Params
    ------
    hcapi: The api instance to run commands through
    mount: The mount to calibrate
    slot_edge_nominal: The point describing the nominal position of the
        edge that we're checking. Its in-axis coordinate (i.e. its x coordinate
        for an x edge) should be the nominal position that we'll compare to. Its
        cross-axis coordiante (i.e. its y coordinate for an x edge) should be
        the point along the edge to search at, usually the midpoint. Its z-axis
        coordinate should be the current best estimate for the height of the deck.
    search_axis: The axis along which to search
    direction_if_hit: The direction to search next if the probe hits the deck.
    Returns
    -------
    The absolute position at which the center of the effector is inside the slot
    and its edge is aligned with the calibration slot edge.
    """
    edge_settings = hcapi.config.calibration.edge_sense
    # Our first search position is at the slot edge nominal at the probe's edge
    checking_pos = slot_edge_nominal + critical_edge_offset(
        search_axis, direction_if_hit
    )
    stride = edge_settings.search_initial_tolerance_mm * direction_if_hit
    final_z_height_found = slot_edge_nominal.z
    for _ in range(edge_settings.search_iteration_limit):
        LOG.info(f"Checking position {checking_pos}")
        interaction_pos, hit_deck = await _probe_deck_at(
            hcapi, mount, checking_pos, edge_settings.pass_settings, probe=probe
        )
        _verify_height(interaction_pos, checking_pos.z, edge_settings)
        if hit_deck:
            # In this block, we've hit the deck
            LOG.info(f"hit at {interaction_pos}, stride size: {stride}")
            # store the final found Z height found
            # because the height is most accurate next to the edge
            final_z_height_found = interaction_pos
            if copysign(stride, direction_if_hit) == stride:
                # If we're in direction_if_hit direction, the last probe was on the deck,
                # so we want to continue
                stride = stride / 2
            else:
                # if we're against direction_if_hit, the last probe missed,
                # so we want to switch back to narrow things down
                stride = -stride / 2
        else:
            LOG.info(f"Miss at {interaction_pos}, stride size: {stride}")
            # In this block, we've missed the deck
            if copysign(stride, direction_if_hit) == stride:
                # if we are moving in the same direction as direction_if_hit, then we
                # need to reverse - this would be the first time we've missed the deck
                stride = -stride / 2
            else:
                # if we are against direction_if_hit, the last test was off the
                # deck too, so we want to continue
                stride = stride / 2
        checking_pos += search_axis.set_in_point(Point(0, 0, 0), stride)

    try:
        await _verify_edge_pos(
            hcapi,
            mount,
            search_axis,
            checking_pos,
            abs(stride * 2),
            direction_if_hit,
            probe=probe,
        )
    except EdgeNotFoundError as e:
        if raise_verify_error:
            raise
        else:
            LOG.warning(e)
    # remove probe offset so we actually get position of the edge
    found_edge = checking_pos - critical_edge_offset(search_axis, direction_if_hit)
    # use the last-found Z height as the edge's most true Z position
    found_edge = found_edge._replace(z=final_z_height_found)
    return found_edge


async def find_slot_center_binary(
    hcapi: OT3HardwareControlAPI,
    mount: OT3Mount,
    estimated_center: Point,
    raise_verify_error: bool = True,
    probe: InstrumentProbeType = InstrumentProbeType.PRIMARY,
) -> Point:
    """Find the center of the calibration slot by binary-searching its edges.
    Returns the XY-center of the slot.
    """
    # Find all four edges of the calibration slot
    plus_x_edge = await find_edge_binary(
        hcapi,
        mount,
        estimated_center + EDGES["right"],
        Axis.X,
        -1,
        raise_verify_error,
        probe=probe,
    )
    LOG.info(f"Found +x edge at {plus_x_edge.x}mm")
    estimated_center = estimated_center._replace(x=plus_x_edge.x - EDGES["right"].x)

    plus_y_edge = await find_edge_binary(
        hcapi,
        mount,
        estimated_center + EDGES["top"],
        Axis.Y,
        -1,
        raise_verify_error,
        probe=probe,
    )
    LOG.info(f"Found +y edge at {plus_y_edge.y}mm")
    estimated_center = estimated_center._replace(y=plus_y_edge.y - EDGES["top"].y)

    minus_x_edge = await find_edge_binary(
        hcapi,
        mount,
        estimated_center + EDGES["left"],
        Axis.X,
        1,
        raise_verify_error,
        probe=probe,
    )
    LOG.info(f"Found -x edge at {minus_x_edge.x}mm")
    estimated_center = estimated_center._replace(x=(plus_x_edge.x + minus_x_edge.x) / 2)

    minus_y_edge = await find_edge_binary(
        hcapi,
        mount,
        estimated_center + EDGES["bottom"],
        Axis.Y,
        1,
        raise_verify_error,
        probe=probe,
    )
    LOG.info(f"Found -y edge at {minus_y_edge.y}mm")
    estimated_center = estimated_center._replace(y=(plus_y_edge.y + minus_y_edge.y) / 2)

    # Found XY center and the average of the edges' Zs
    return estimated_center._replace(
        z=(plus_x_edge.z + minus_x_edge.z + plus_y_edge.z + minus_y_edge.z) / 4,
    )


async def find_calibration_structure_height(
    hcapi: OT3HardwareControlAPI,
    mount: OT3Mount,
    nominal_center: Point,
    probe: InstrumentProbeType = InstrumentProbeType.PRIMARY,
) -> float:
    """
    Find the height of the calibration structure in this mount's frame of reference.

    This could be the deck height or the module calibration adapter height.

    The deck nominal height in deck coordinates is 0 (that's part of the
    definition of deck coordinates) but if we have not yet calibrated a
    particular tool on a particular mount, then the z deck coordinate that
    will cause a collision is not 0. This routine finds that value.
    """
    z_pass_settings = hcapi.config.calibration.z_offset.pass_settings
    z_prep_point = nominal_center + PREP_OFFSET_DEPTH
    z_limit = nominal_center.z - z_pass_settings.max_overrun_distance_mm
    structure_z, hit_deck = await _probe_deck_at(
        hcapi, mount, z_prep_point, z_pass_settings, probe=probe
    )
    if not hit_deck:
        raise CalibrationStructureNotFoundError(structure_z, z_limit)
    LOG.info(f"autocalibration: found structure at {structure_z}")
    return structure_z


async def _probe_deck_at(
    api: OT3HardwareControlAPI,
    mount: OT3Mount,
    target: Point,
    settings: CapacitivePassSettings,
    speed: float = 50,
    probe: InstrumentProbeType = InstrumentProbeType.PRIMARY,
) -> Tuple[float, bool]:
    here = await api.gantry_position(mount)
    abs_transit_height = max(
        target.z + LINEAR_TRANSIT_HEIGHT, target.z + settings.prep_distance_mm
    )
    safe_height = max(here.z, target.z, abs_transit_height)
    await api.move_to(mount, here._replace(z=safe_height))
    await api.move_to(mount, target._replace(z=safe_height), speed=speed)
    await api.move_to(mount, target._replace(z=abs_transit_height))
    _found_pos, contact = await api.capacitive_probe(
        mount, Axis.by_mount(mount), target.z, settings, probe=probe
    )
    # don't use found Z position to calculate an updated transit height
    # because the probe may have gone through the hole
    await api.move_to(mount, target._replace(z=abs_transit_height))
    return _found_pos, contact


async def find_axis_center(
    hcapi: OT3HardwareControlAPI,
    mount: OT3Mount,
    minus_edge_nominal: Point,
    plus_edge_nominal: Point,
    axis: Literal[Axis.X, Axis.Y],
) -> float:
    """Find the center of the calibration slot on the specified axis.

    Sweep from the specified left edge to the specified right edge while taking
    capacitive sense data. When the probe is over the deck, the capacitance will
    be higher than when the probe is over the slot. By postprocessing the data,
    we determine where the slot edges are, and return those positions.
    """
    WIDTH_TOLERANCE_MM: float = 0.5
    here = await hcapi.gantry_position(mount)
    await hcapi.move_to(mount, here._replace(z=SEARCH_TRANSIT_HEIGHT))
    edge_settings = hcapi.config.calibration.edge_sense

    start = axis.set_in_point(
        minus_edge_nominal,
        axis.of_point(minus_edge_nominal) - edge_settings.search_initial_tolerance_mm,
    )
    end = axis.set_in_point(
        plus_edge_nominal,
        axis.of_point(plus_edge_nominal) + edge_settings.search_initial_tolerance_mm,
    )

    await hcapi.move_to(mount, start._replace(z=SEARCH_TRANSIT_HEIGHT))

    data = await hcapi.capacitive_sweep(
        mount, axis, start, end, edge_settings.pass_settings.speed_mm_per_s
    )

    left_edge, right_edge = _edges_from_data(
        data,
        axis.of_point(end) - axis.of_point(start),
        {
            "axis": axis.name,
            "speed": edge_settings.pass_settings.speed_mm_per_s,
            "start_absolute": axis.of_point(start),
            "end_absolute": axis.of_point(end),
        },
    )
    nominal_width = axis.of_point(plus_edge_nominal) - axis.of_point(minus_edge_nominal)
    detected_width = right_edge - left_edge
    left_edge_absolute = axis.of_point(start) + left_edge
    right_edge_absolute = axis.of_point(start) + right_edge
    if abs(detected_width - nominal_width) > WIDTH_TOLERANCE_MM:
        raise InaccurateNonContactSweepError(nominal_width, detected_width)
    return (left_edge_absolute + right_edge_absolute) / 2


def _edges_from_data(
    data: List[float], distance: float, log_metadata: Optional[Dict[str, Any]] = None
) -> Tuple[float, float]:
    """
    Postprocess the capacitance data taken from a sweep to find the calibration slot.

    The sweep should have covered both edges, going off the deck into the slot,
    all the way across the slot, and back onto the deck on the other side.

    Capacitance is proportional to the area of the "plates" involved in the sensing. To
    a first approximation, these are the flat circular face of the bottom of the probe,
    and the deck. When the probe begins to cross the edge of the deck, the area of the
    second "plate" - the deck - ends abruptly, in a straight line transverse to motion.
    As the probe crosses, less and less of that circular face is over the deck.

    That means that the rate of change with respect to the overlap distance (or time, at
    constant velocity) is maximized as the center of the probe passes the edge of the
    deck.

    We can therefore apply a combined smoothing and differencing convolution kernel to
    the timeseries data and set the locations of the edges as the locations of the
    extrema of the difference of the series.
    """
    now_str = datetime.datetime.now().strftime("%d-%m-%y-%H:%M:%S")

    # The width of the averaging kernel defines how strong the averaging is - a wider
    # kernel, or filter, has a lower rolloff frequency and will smooth more. This
    # calculation sets the width at 5% of the length of the data, and then makes that
    # value an even number
    average_width_samples = (int(floor(0.05 * len(data))) // 2) * 2
    # an averaging kernel would be an array of length N with elements each set to 1/N;
    # when convolved with a data stream, this will (ignoring edge effects) produce
    # an N-sample rolling average. by inverting the sign of half the kernel, which is
    # why we need it to be even, we do the same thing but while also taking a finite
    # difference.
    average_difference_kernel = np.concatenate(
        (
            np.full(average_width_samples // 2, 1 / average_width_samples),
            np.full(average_width_samples // 2, -1 / average_width_samples),
        )
    )
    differenced = np.convolve(np.array(data), average_difference_kernel, mode="valid")
    # These are the indices of the minimum difference (which should be the left edge,
    # where the probe is halfway through moving off the deck, and the slope of the
    # data is most negative) and the maximum difference (which should be the right
    # edge, where the probe is halfway through moving back onto the deck, and the slope
    # of the data is most positive)
    left_edge_sample = np.argmin(differenced)
    right_edge_sample = np.argmax(differenced)
    mm_per_elem = distance / len(data)
    # The differenced data is shorter than the input data because we used valid outputs
    # of the convolution only to avoid edge effects; that means we need to account for
    # the distance in the cut-off data
    distance_prefix = ((len(data) - len(differenced)) / 2) * mm_per_elem
    left_edge_offset = left_edge_sample * mm_per_elem
    left_edge = left_edge_offset + distance_prefix
    right_edge_offset = right_edge_sample * mm_per_elem
    right_edge = right_edge_offset + distance_prefix
    json.dump(
        {
            "metadata": log_metadata,
            "inputs": {
                "raw_data": data,
                "distance": distance,
                "kernel_width": len(average_difference_kernel),
            },
            "outputs": {
                "differenced_data": [d for d in differenced],
                "mm_per_elem": mm_per_elem,
                "distance_prefix": distance_prefix,
                "right_edge_offset": right_edge_offset,
                "left_edge_offset": left_edge_offset,
                "left_edge": left_edge,
                "right_edge": right_edge,
            },
        },
        open(f"/data/sweep_{now_str}.json", "w"),
    )
    LOG.info(
        f"Found edges ({left_edge:.3f}, {right_edge:.3f}) "
        f"from offsets ({left_edge_offset:.3f}, {right_edge_offset:.3f}) "
        f"with {len(data)} cap samples over {distance}mm "
        f"using a kernel width of {len(average_difference_kernel)}"
    )
    return float(left_edge), float(right_edge)


async def find_slot_center_noncontact(
    hcapi: OT3HardwareControlAPI, mount: OT3Mount, estimated_center: Point
) -> Point:
    NONCONTACT_INTERVAL_MM: float = 0.1
    travel_center = estimated_center + Point(0, 0, NONCONTACT_INTERVAL_MM)
    x_center = await find_axis_center(
        hcapi,
        mount,
        travel_center + EDGES["left"],
        travel_center + EDGES["right"],
        Axis.X,
    )
    y_center = await find_axis_center(
        hcapi,
        mount,
        travel_center + EDGES["bottom"],
        travel_center + EDGES["top"],
        Axis.Y,
    )
    return Point(x_center, y_center, estimated_center.z)


async def find_calibration_structure_center(
    hcapi: OT3HardwareControlAPI,
    mount: OT3Mount,
    nominal_center: Point,
    method: CalibrationMethod = CalibrationMethod.BINARY_SEARCH,
    raise_verify_error: bool = True,
    probe: InstrumentProbeType = InstrumentProbeType.PRIMARY,
) -> Point:

    # Perform xy offset search
    if method == CalibrationMethod.BINARY_SEARCH:
        found_center = await find_slot_center_binary(
            hcapi, mount, nominal_center, raise_verify_error, probe=probe
        )
    elif method == CalibrationMethod.NONCONTACT_PASS:
        # FIXME: use slot to find ideal position
        found_center = await find_slot_center_noncontact(hcapi, mount, nominal_center)
    else:
        raise RuntimeError("Unknown calibration method")
    return found_center


async def _calibrate_mount(
    hcapi: OT3HardwareControlAPI,
    mount: OT3Mount,
    slot: int = SLOT_CENTER,
    method: CalibrationMethod = CalibrationMethod.BINARY_SEARCH,
    raise_verify_error: bool = True,
    probe: InstrumentProbeType = InstrumentProbeType.PRIMARY,
) -> Point:
    """
    Run automatic calibration for the tool attached to the specified mount.

    Before running this function, make sure that the appropriate probe
    has been attached or prepped on the tool (for instance, a capacitive
    tip has been attached, or the conductive probe has been attached,
    or the probe has been lowered). The robot should be homed.

    Note: To calibrate a gripper, this process must be performed on the front
    and rear calibration pins separately. The gripper calibration offset is
    the average of the pin offsets, which can be obtained by passing the
    two offsets into the `gripper_pin_offsets_mean` func.

    Params
    ------
    hcapi: a hardware control api to run commands against
    mount: The mount to calibration

    Returns
    -------
    The estimated position of the XY center of the calibration slot in
    the plane of the deck. This value is suitable for vector-subtracting
    from the current instrument offset to set a new instrument offset.
    """
    nominal_center = Point(*get_calibration_square_position_in_slot(slot))
    if probe == InstrumentProbeType.SECONDARY and mount != OT3Mount.GRIPPER:
        pip = hcapi.hardware_instruments[mount.to_mount()]
        num_channels = int(pip.channels)  # type: ignore[union-attr]
        nominal_center += OFFSET_SECONDARY_PROBE.get(num_channels, Point())
    async with hcapi.restore_system_constrants():
        await hcapi.set_system_constraints_for_calibration()
        try:
            # find the center of the calibration sqaure
            offset = await find_calibration_structure_position(
                hcapi,
                mount,
                nominal_center,
                method=method,
                raise_verify_error=raise_verify_error,
                probe=probe,
            )
            # update center with values obtained during calibration
            LOG.info(f"Found calibration value {offset} for mount {mount.name}")
            return offset

        except (
            InaccurateNonContactSweepError,
            EarlyCapacitiveSenseTrigger,
            CalibrationStructureNotFoundError,
            EdgeNotFoundError,
        ):
            LOG.info(
                "Error occurred during calibration. Resetting to current saved calibration value."
            )
            await hcapi.reset_instrument_offset(mount, to_default=False)
            # re-raise exception after resetting instrument offset
            raise


async def find_calibration_structure_position(
    hcapi: OT3HardwareControlAPI,
    mount: OT3Mount,
    nominal_center: Point,
    method: CalibrationMethod = CalibrationMethod.BINARY_SEARCH,
    target: CalibrationTarget = CalibrationTarget.GANTRY_INSTRUMENT,
    raise_verify_error: bool = True,
    probe: InstrumentProbeType = InstrumentProbeType.PRIMARY,
) -> Point:
    """Find the calibration square offset given an arbitry postition on the deck."""
    # Find the estimated structure plate height. This will be used to baseline the edge detection points.
    z_height = await find_calibration_structure_height(
        hcapi, mount, nominal_center, probe=probe
    )
    initial_center = nominal_center._replace(z=z_height)
    LOG.info(f"Found structure plate at {z_height}mm")

    # Find the calibration square center using the given method
    found_center = await find_calibration_structure_center(
        hcapi, mount, initial_center, method, raise_verify_error, probe=probe
    )

    offset = nominal_center - found_center
    # NOTE: If the calibration target is a deck object the polarity of the calibrated
    #  offset needs to be reversed. This is because we are using the gantry instrument
    #  to calibrate a stationary object on the deck and need to find the offset of that
    #  deck object relative to the deck and not the instrument which sits above the deck.
    if target == CalibrationTarget.DECK_OBJECT:
        return offset * -1
    return offset


async def find_slot_center_binary_from_nominal_center(
    hcapi: OT3HardwareControlAPI,
    mount: OT3Mount,
    slot: int,
) -> Tuple[Point, Point]:
    """
    For use with calibrate_belts. For specified slot, finds actual slot center via binary search and nominal slot center

    Params
    ------
    hcapi: a hardware control api to run commands against
    mount: the mount to calibration
    slot: a specific deck slot

    Returns
    -------
    The actual and nominal centers of the specified slot.
    """
    nominal_center = Point(*get_calibration_square_position_in_slot(slot))
    offset = await find_calibration_structure_position(
        hcapi, mount, nominal_center, method=CalibrationMethod.BINARY_SEARCH
    )
    return nominal_center - offset, nominal_center


async def _determine_transform_matrix(
    hcapi: OT3HardwareControlAPI,
    mount: OT3Mount,
) -> Tuple[types.AttitudeMatrix, Dict[str, Any]]:
    """
    Run automatic calibration for the gantry x and y belts attached to the specified mount. Returned linear transform matrix is determined via the
    actual and nominal center points of the back right (A), front right (B), and back left (C) slots.

    Params
    ------
    hcapi: a hardware control api to run commands against
    mount: the mount to calibration

    Returns
    -------
    A listed matrix of the linear transform in the x and y dimensions that accounts for the stretch of the gantry x and y belts.
    """

    async def _find_slot(s: int) -> CalibrationSlot:
        actual, nominal = await find_slot_center_binary_from_nominal_center(
            hcapi, mount, s
        )
        await hcapi.retract(mount)
        return CalibrationSlot(slot=s, nominal=nominal, actual=actual)

    front_left = await _find_slot(SLOT_FRONT_LEFT)
    front_right = await _find_slot(SLOT_FRONT_RIGHT)
    rear_left = await _find_slot(SLOT_REAR_LEFT)
    belt_cal = BeltCalibrationData(front_left, front_right, rear_left)
    details = belt_cal.check_alignment()  # raises error if misaligned
    attitude = solve_attitude(*belt_cal.get_solve_points())
    return attitude, details


def gripper_pin_offsets_mean(front: Point, rear: Point) -> Point:
    """
    Get calibration offset of a gripper from its front and rear pin offsets.

    This function should be used for gripper calibration only.

    Params
    ------
    front: gripper's front pin calibration offset
    rear: gripper's rear pin calibration offset

    Returns
    -------
    The gripper calibration offset.
    """
    return 0.5 * (front + rear)


async def calibrate_gripper_jaw(
    hcapi: OT3HardwareControlAPI,
    probe: GripperProbe,
    slot: int = 5,
    method: CalibrationMethod = CalibrationMethod.BINARY_SEARCH,
    raise_verify_error: bool = True,
) -> Point:
    """
    Run automatic calibration for gripper jaw.

    Before running this function, make sure that the appropriate probe
    has been attached or prepped on the tool (for instance, a capacitive
    tip has been attached, or the conductive probe has been attached,
    or the probe has been lowered). The robot should be homed.

    This process must be performed on the front
    and rear calibration pins separately. The gripper calibration offset is
    the average of the pin offsets, which can be obtained by passing the
    two offsets into the `gripper_pin_offsets_mean` func.
    """
    try:
        await hcapi.reset_instrument_offset(OT3Mount.GRIPPER)
        hcapi.add_gripper_probe(probe)
        await hcapi.grip(GRIPPER_GRIP_FORCE)
        offset = await _calibrate_mount(
            hcapi,
            OT3Mount.GRIPPER,
            slot,
            method,
            raise_verify_error,
            probe=probe.to_type(probe),
        )
        LOG.info(f"Gripper {probe.name} probe offset: {offset}")
        return offset
    finally:
        hcapi.remove_gripper_probe()


async def calibrate_gripper(
    hcapi: OT3HardwareControlAPI, offset_front: Point, offset_rear: Point
) -> Point:
    """Calibrate gripper."""
    offset = gripper_pin_offsets_mean(front=offset_front, rear=offset_rear)
    LOG.info(f"Gripper calibration offset: {offset}")
    await hcapi.save_instrument_offset(OT3Mount.GRIPPER, offset)
    return offset


async def find_pipette_offset(
    hcapi: OT3HardwareControlAPI,
    mount: Literal[OT3Mount.LEFT, OT3Mount.RIGHT],
    slot: int = 5,
    method: CalibrationMethod = CalibrationMethod.BINARY_SEARCH,
    raise_verify_error: bool = True,
    reset_instrument_offset: bool = True,
    probe: InstrumentProbeType = InstrumentProbeType.PRIMARY,
) -> Point:
    """
    Run automatic calibration for pipette and only return the calibration point.

    Before running this function, make sure that the appropriate probe
    has been attached or prepped on the tool (for instance, a capacitive
    tip has been attached, or the conductive probe has been attached,
    or the probe has been lowered).

    This function should be used in the robot server only.
    """
    try:
        if reset_instrument_offset:
            await hcapi.reset_instrument_offset(mount)
        hcapi.add_tip(mount, hcapi.config.calibration.probe_length)
        offset = await _calibrate_mount(
            hcapi, mount, slot, method, raise_verify_error, probe=probe
        )
        return offset
    finally:
        hcapi.remove_tip(mount)


async def calibrate_pipette(
    hcapi: OT3HardwareControlAPI,
    mount: Literal[OT3Mount.LEFT, OT3Mount.RIGHT],
    slot: int = 5,
    method: CalibrationMethod = CalibrationMethod.BINARY_SEARCH,
    raise_verify_error: bool = True,
    probe: InstrumentProbeType = InstrumentProbeType.PRIMARY,
) -> Point:
    """
    Run automatic calibration for pipette and save the offset.

    Before running this function, make sure that the appropriate probe
    has been attached or prepped on the tool (for instance, a capacitive
    tip has been attached, or the conductive probe has been attached,
    or the probe has been lowered).
    """
    offset = await find_pipette_offset(
        hcapi, mount, slot, method, raise_verify_error, probe=probe
    )
    await hcapi.save_instrument_offset(mount, offset)
    return offset


async def calibrate_module(
    hcapi: OT3HardwareControlAPI,
    mount: OT3Mount,
    slot: str,
    module_id: str,
    nominal_position: Point,
) -> Point:
    """
    Run automatic calibration for a module.

    Before running this function, make sure that the appropriate probe
    has been attached or prepped on the tool (for instance, a capacitive
    tip has been attached, or the conductive probe has been attached,
    or the probe has been lowered). We also need to have the module
    prepped for calibration (for example, not heating, lid closed,
    not shaking, etc) and the corresponding module calibration
    block placed in the module slot.

    The robot should be homed before calling this function.
    """

    try:
        # add the probe depending on the mount
        if mount == OT3Mount.GRIPPER:
            hcapi.add_gripper_probe(GripperProbe.FRONT)
        else:
            hcapi.add_tip(mount, hcapi.config.calibration.probe_length)

        LOG.info(
            f"Starting module calibration for {module_id} at {nominal_position} using {mount}"
        )
        # FIXME (ba, 2023-04-04): Well B1 of the module adapter definition includes the z prep offset
        # of 13x13mm in the nominial position, but we are still using PREP_OFFSET_DEPTH in
        # find_calibration_structure_height which effectively doubles the offset. We plan
        # on removing PREP_OFFSET_DEPTH in the near future, but for now just subtract PREP_OFFSET_DEPTH
        # from the nominal position so we dont have to alter any other part of the system.
        nominal_position = nominal_position - PREP_OFFSET_DEPTH
        offset = await find_calibration_structure_position(
            hcapi,
            mount,
            nominal_position,
            method=CalibrationMethod.BINARY_SEARCH,
            target=CalibrationTarget.DECK_OBJECT,
        )
        await hcapi.save_module_offset(module_id, mount, slot, offset)
        return offset
    finally:
        # remove probe
        if mount == OT3Mount.GRIPPER:
            hcapi.remove_gripper_probe()
            await hcapi.ungrip()
        else:
            hcapi.remove_tip(mount)


async def calibrate_belts(
    hcapi: OT3HardwareControlAPI,
    mount: OT3Mount,
    pipette_id: str,
) -> Tuple[types.AttitudeMatrix, Dict[str, Any]]:
    """
    Run automatic calibration for the gantry x and y belts attached to the specified mount.

    Params
    ------
    hcapi: a hardware control api to run commands against
    mount: the mount to calibration

    Returns
    -------
    A listed matrix of the linear transform in the x and y dimensions that accounts for the stretch of the gantry x and y belts.
    """
    if mount == OT3Mount.GRIPPER:
        raise RuntimeError("Must use pipette mount, not gripper")
    try:
        hcapi.reset_deck_calibration()
        hcapi.add_tip(mount, hcapi.config.calibration.probe_length)
        belt_attitude, alignment_details = await _determine_transform_matrix(
            hcapi, mount
        )
        save_robot_belt_attitude(belt_attitude, pipette_id)
        return belt_attitude, alignment_details
    finally:
        hcapi.load_deck_calibration()
        hcapi.remove_tip(mount)


def apply_machine_transform(
    belt_attitude: types.AttitudeMatrix,
) -> types.AttitudeMatrix:
    """
    This applies the machine attitude matrix (which happens to be a negative identity) to the belt attitude matrix to form the deck attitude matrix.

    Param
    -----
    belt_attitude: attitude matrix with regards to belt coordinate system

    Returns
    -------
    Attitude matrix with regards to machine coordinate system.
    """
    belt_attitude_arr: DoubleMatrix = np.array(belt_attitude)
    machine_transform_arr: DoubleMatrix = np.array(
        defaults_ot3.DEFAULT_MACHINE_TRANSFORM
    )
    deck_attitude_arr = np.dot(belt_attitude_arr, machine_transform_arr)
    deck_attitude = deck_attitude_arr.round(4).tolist()
    return deck_attitude  # type: ignore[no-any-return]


def load_attitude_matrix(to_default: bool = True) -> DeckCalibration:
    calibration_data = get_robot_belt_attitude()

    if calibration_data and not to_default:
        return DeckCalibration(
            attitude=apply_machine_transform(calibration_data.attitude),
            source=calibration_data.source,
            status=types.CalibrationStatus(**calibration_data.status.model_dump()),
            belt_attitude=calibration_data.attitude,
            last_modified=calibration_data.lastModified,
            pipette_calibrated_with=calibration_data.pipetteCalibratedWith,
        )
    else:
        # load default if calibration data does not exist
        return DeckCalibration(
            attitude=apply_machine_transform(default_ot3_deck_calibration()),
            source=types.SourceType.default,
            status=types.CalibrationStatus(),
            belt_attitude=default_ot3_deck_calibration(),
        )


def validate_attitude_deck_calibration(
    deck_cal: DeckCalibration,
) -> DeckTransformState:
    """
    This function determines whether the deck calibration is valid
    or not based on the following use-cases:

    TODO(pm, 5/9/2023): As with the OT2, expand on this method,
    or create another method to diagnose bad instrument offset data
    """
    curr_cal: DoubleMatrix = np.array(deck_cal.attitude)
    row, _ = curr_cal.shape
    rank: int = np.linalg.matrix_rank(curr_cal)
    if row != rank:
        # Check that the matrix is non-singular
        return DeckTransformState.SINGULARITY
    elif not deck_cal.last_modified:
        # Check that the matrix is not an identity
        return DeckTransformState.IDENTITY
    else:
        # Transform as it stands is sufficient.
        return DeckTransformState.OK


def delete_belt_calibration_data(hcapi: OT3HardwareControlAPI) -> None:
    delete_robot_belt_attitude()
    hcapi.reset_deck_calibration()


class OT3RobotCalibrationProvider:
    """This class provides the following robot calibration data:
    deck calibration: transform matrix to account for stretch of x and y belts
    carriage offset: the vector from the deck origin to the bottom center of the gantry carriage when the gantry is homed
    mount offset (per mount): the vector from the carriage origin (bottom center) to the mount origin (centered on the top peg of the mount flush with the mating face)
    """

    def __init__(self, config: OT3Config) -> None:
        self._robot_calibration = OT3Transforms(
            deck_calibration=load_attitude_matrix(to_default=False),
            carriage_offset=Point(*config.carriage_offset),
            left_mount_offset=Point(*config.left_mount_offset),
            right_mount_offset=Point(*config.right_mount_offset),
            gripper_mount_offset=Point(*config.gripper_mount_offset),
        )

    @lru_cache(1)
    def _validate(self) -> DeckTransformState:
        return validate_attitude_deck_calibration(
            self._robot_calibration.deck_calibration
        )

    @property
    def robot_calibration(self) -> OT3Transforms:
        return self._robot_calibration

    def reset_robot_calibration(self) -> None:
        self._validate.cache_clear()
        self._robot_calibration = OT3Transforms(
            deck_calibration=load_attitude_matrix(to_default=True),
            carriage_offset=Point(*defaults_ot3.DEFAULT_CARRIAGE_OFFSET),
            left_mount_offset=Point(*defaults_ot3.DEFAULT_LEFT_MOUNT_OFFSET),
            right_mount_offset=Point(*defaults_ot3.DEFAULT_RIGHT_MOUNT_OFFSET),
            gripper_mount_offset=Point(*defaults_ot3.DEFAULT_GRIPPER_MOUNT_OFFSET),
        )

    def reset_deck_calibration(self) -> None:
        self._robot_calibration.deck_calibration = load_attitude_matrix(to_default=True)

    def load_deck_calibration(self) -> None:
        self._validate.cache_clear()
        self._robot_calibration.deck_calibration = load_attitude_matrix(
            to_default=False
        )

    def set_robot_calibration(self, robot_calibration: OT3Transforms) -> None:
        self._validate.cache_clear()
        self._robot_calibration = robot_calibration

    def validate_calibration(self) -> DeckTransformState:
        """
        The lru cache decorator is currently not supported by the
        ThreadManager. To work around this, we need to wrap the
        actual function around a dummy outer function.

        Once decorators are more fully supported, we can remove this.
        """
        return self._validate()

    def build_temporary_identity_calibration(self) -> OT3Transforms:
        """
        Get temporary default calibration data suitable for use during
        calibration
        """
        return OT3Transforms(
            deck_calibration=load_attitude_matrix(to_default=True),
            carriage_offset=Point(*defaults_ot3.DEFAULT_CARRIAGE_OFFSET),
            left_mount_offset=Point(*defaults_ot3.DEFAULT_LEFT_MOUNT_OFFSET),
            right_mount_offset=Point(*defaults_ot3.DEFAULT_RIGHT_MOUNT_OFFSET),
            gripper_mount_offset=Point(*defaults_ot3.DEFAULT_GRIPPER_MOUNT_OFFSET),
        )


@dataclass
class OT3Transforms(RobotCalibration):
    carriage_offset: Point
    left_mount_offset: Point
    right_mount_offset: Point
    gripper_mount_offset: Point


def _point_to_tuple(_p: Point) -> Tuple[float, float, float]:
    return _p.x, _p.y, _p.z


class BeltCalibrationData:
    def __init__(
        self,
        slot_front_left: CalibrationSlot,
        slot_front_right: CalibrationSlot,
        slot_rear_left: CalibrationSlot,
    ) -> None:
        self._front_left = slot_front_left
        self._front_right = slot_front_right
        self._rear_left = slot_rear_left

    def build_details(self) -> Dict[str, Any]:
        shift_details = {
            shift.value: {
                "spec": MAX_SHIFT[shift],
                "pass": abs(self._get_shift_mm(shift)) < MAX_SHIFT[shift],
                "shift": round(self._get_shift_mm(shift), 3),
            }
            for shift in AlignmentShift
        }
        shift_details["slots"] = {
            "front_left": _point_to_tuple(self._front_left.actual),  # type: ignore[dict-item]
            "front_right": _point_to_tuple(self._front_right.actual),  # type: ignore[dict-item]
            "rear_left": _point_to_tuple(self._rear_left.actual),  # type: ignore[dict-item]
        }
        return shift_details

    def check_alignment(self) -> Dict[str, Any]:
        shift_details = self.build_details()
        LOG.info(shift_details)
        failures = [
            shift for shift in AlignmentShift if not shift_details[shift.value]["pass"]
        ]
        if failures:
            raise MisalignedGantryError(shift_details)
        return shift_details

    def get_solve_points(self) -> Tuple[SolvePoints, SolvePoints]:
        actual = (
            _point_to_tuple(self._front_left.actual),
            _point_to_tuple(self._rear_left.actual),
            _point_to_tuple(self._front_right.actual),
        )
        nominal = (
            _point_to_tuple(self._front_left.nominal),
            _point_to_tuple(self._rear_left.nominal),
            _point_to_tuple(self._front_right.nominal),
        )
        return nominal, actual

    def _get_shift_mm(self, shift: AlignmentShift) -> float:
        # polarity is same as deck coordinates,
        # so positive values describe shifting towards right/rear/up,
        # while negative values describe shifting towards left/front/down.
        if shift == AlignmentShift.FRONT_TO_REAR_X:
            return self._rear_left.actual.x - self._front_left.actual.x
        elif shift == AlignmentShift.FRONT_TO_REAR_Z:
            return self._rear_left.actual.z - self._front_left.actual.z
        elif shift == AlignmentShift.LEFT_TO_RIGHT_Y:
            return self._front_right.actual.y - self._front_left.actual.y
        elif shift == AlignmentShift.LEFT_TO_RIGHT_Z:
            return self._front_right.actual.z - self._front_left.actual.z
        raise ValueError(f"unexpected shift: {shift}")



""" Utility functions and classes for the hardware controller"""
import asyncio
import logging
from enum import Enum
from typing import Dict, Any, Optional, List, Mapping, Tuple, TypeVar, Union

from .types import CriticalPoint, MotionChecks, Axis
from .errors import OutOfBoundsMove
from opentrons_shared_data.errors.exceptions import MissingConfigurationData
from opentrons.types import Point
from opentrons_shared_data.pipette.types import PipetteTipType
from opentrons_shared_data.pipette.pipette_definition import (
    PipetteConfigurations,
    PressFitPickUpTipConfiguration,
    CamActionPickUpTipConfiguration,
    PressAndCamConfigurationValues,
)

mod_log = logging.getLogger(__name__)


def _handle_loop_exception(
    loop: asyncio.AbstractEventLoop, context: Dict[str, Any]
) -> None:
    mod_log.error(f"Caught exception: {context}")


def use_or_initialize_loop(
    loop: Optional[asyncio.AbstractEventLoop],
) -> asyncio.AbstractEventLoop:
    checked_loop = loop or asyncio.get_event_loop()
    checked_loop.set_exception_handler(_handle_loop_exception)
    return checked_loop


def plan_arc(
    origin_point: Point,
    dest_point: Point,
    z_height: float,
    origin_cp: Optional[CriticalPoint] = None,
    dest_cp: Optional[CriticalPoint] = None,
    extra_waypoints: Optional[List[Tuple[float, float]]] = None,
) -> List[Tuple[Point, Optional[CriticalPoint]]]:

    assert z_height >= max(origin_point.z, dest_point.z)
    checked_wp = extra_waypoints or []
    return (
        [(origin_point._replace(z=z_height), origin_cp)]
        + [(Point(x=wp[0], y=wp[1], z=z_height), dest_cp) for wp in checked_wp]
        + [(dest_point._replace(z=z_height), dest_cp), (dest_point, dest_cp)]
    )


class DeckTransformState(Enum):
    OK = "OK"
    IDENTITY = "IDENTITY"
    BAD_CALIBRATION = "BAD_CALIBRATION"
    SINGULARITY = "SINGULARITY"

    def __str__(self) -> str:
        return self.name


AxisType = TypeVar("AxisType")


def check_motion_bounds(
    target_smoothie: Mapping[AxisType, float],
    target_deck: Mapping[AxisType, float],
    bounds: Mapping[AxisType, Tuple[float, float]],
    checks: MotionChecks,
) -> None:
    """
    Log or raise an error (depending on checks) if a specified
    target position is outside the acceptable bounds of motion

    Which axes are checked is defined by the elements of
    target_smoothie.

    The target_deck and mapping does not need to be
    entirely filled; it is used only for logging
    """
    bounds_message_format = (
        "Out of bounds move: {axis}=({tsp} motor controller, {tdp} deck) too "
        "{dir} for limit {limsp}"
    )
    for ax in target_smoothie.keys():
        if target_smoothie[ax] < bounds[ax][0]:
            format_detail = {
                "axis": ax,
                "tsp": target_smoothie[ax],
                "tdp": target_deck.get(ax, "unknown"),
                "dir": "low",
                "limsp": bounds.get(ax, ("unknown",))[0],
            }
            bounds_message = bounds_message_format.format(**format_detail)
            mod_log.warning(bounds_message)
            if checks.value & MotionChecks.LOW.value:
                raise OutOfBoundsMove(bounds_message, format_detail)
        elif target_smoothie[ax] > bounds[ax][1]:
            format_detail = {
                "axis": ax,
                "tsp": target_smoothie[ax],
                "tdp": target_deck.get(ax, "unknown"),
                "dir": "high",
                "limsp": bounds.get(ax, (None, "unknown"))[1],
            }
            bounds_message = bounds_message_format.format(**format_detail)
            mod_log.warning(bounds_message)
            if checks.value & MotionChecks.HIGH.value:
                raise OutOfBoundsMove(bounds_message, format_detail)


def ot2_axis_to_string(axis: Axis) -> str:
    """Returns OT-2 specific string for the given axis."""
    axis_str_map = {
        Axis.Z_L: "Z",
        Axis.Z_R: "A",
        Axis.P_L: "B",
        Axis.P_R: "C",
    }
    try:
        return axis_str_map[axis]
    except KeyError:
        return axis.name


def _get_press_and_cam_configuration_values(
    config: Union[CamActionPickUpTipConfiguration, PressFitPickUpTipConfiguration],
    valid_nozzle_map_key: str,
    active_tip_type: PipetteTipType,
) -> PressAndCamConfigurationValues:
    try:
        return config.configuration_by_nozzle_map[valid_nozzle_map_key][
            active_tip_type.name
        ]
    except KeyError:
        default = config.configuration_by_nozzle_map[valid_nozzle_map_key].get(
            "default"
        )
        if default is not None:
            return default
        raise MissingConfigurationData(
            message=f"Default tip type configuration values do not exist for Nozzle Map {valid_nozzle_map_key}."
        )


def pick_up_speed_by_configuration(
    config: Union[CamActionPickUpTipConfiguration, PressFitPickUpTipConfiguration],
    valid_nozzle_map_key: str,
    active_tip_type: PipetteTipType,
) -> float:
    """
    Returns the pick up Speed for a given configuration of a given pipette.

    Parameters:
        config: A PressFitPickUpTipConfiguration or a CamActionPickUpTipConfiguration (96ch Only) containing values for speed.
        valid_nozzle_map_key: The string key representing the current pre-validated map of nozzles the pipette has been configured to use.
        active_tip_type: The PipetteTipType to be attached to the pipette.
    """
    configuration_values = _get_press_and_cam_configuration_values(
        config, valid_nozzle_map_key, active_tip_type
    )
    return configuration_values.speed


def pick_up_distance_by_configuration(
    config: Union[CamActionPickUpTipConfiguration, PressFitPickUpTipConfiguration],
    valid_nozzle_map_key: str,
    active_tip_type: PipetteTipType,
) -> float:
    """
    Returns the pick up Distance for a given configuration of a given pipette.

    Parameters:
        config: A PressFitPickUpTipConfiguration or a CamActionPickUpTipConfiguration (96ch Only) containing values for distance.
        valid_nozzle_map_key: The string key representing the current pre-validated map of nozzles the pipette has been configured to use.
        active_tip_type: The PipetteTipType to be attached to the pipette.
    """
    configuration_values = _get_press_and_cam_configuration_values(
        config, valid_nozzle_map_key, active_tip_type
    )
    return configuration_values.distance


def pick_up_current_by_configuration(
    config: Union[CamActionPickUpTipConfiguration, PressFitPickUpTipConfiguration],
    valid_nozzle_map_key: str,
    active_tip_type: PipetteTipType,
) -> float:
    """
    Returns the pick up Current for a given configuration of a given pipette.

    Parameters:
        config: A PressFitPickUpTipConfiguration or a CamActionPickUpTipConfiguration (96ch Only) containing values for current.
        valid_nozzle_map_key: The string key representing the current pre-validated map of nozzles the pipette has been configured to use.
        active_tip_type: The PipetteTipType to be attached to the pipette.
    """
    configuration_values = _get_press_and_cam_configuration_values(
        config, valid_nozzle_map_key, active_tip_type
    )
    return configuration_values.current


def nominal_tip_overlap_dictionary_by_configuration(
    configurations: PipetteConfigurations,
    valid_nozzle_map_key: str,
    active_tip_type: PipetteTipType,
) -> Dict[str, Dict[str, float]]:
    """
    Returns the pick up speed for a given configuration of a given pipette.

    Parameters:
        configurations: The PipetterConfigurations of the pipette the tips will be attached to.
        valid_nozzle_map_key: The string key representing the current pre-validated map of nozzles the pipette has been configured to use.
        active_tip_type: The PipetteTipType to be attached to the pipette.
    """
    for config in (
        configurations.pick_up_tip_configurations.press_fit,
        configurations.pick_up_tip_configurations.cam_action,
    ):
        if not config:
            continue
        try:
            configuration_values = _get_press_and_cam_configuration_values(
                config, valid_nozzle_map_key, active_tip_type
            )
            return configuration_values.versioned_tip_overlap_dictionary
        except KeyError:
            # No valid key found for the approved nozzle map under this configuration - try the next
            continue
    raise MissingConfigurationData(
        message=f"No valid tip overlap dictionaries identified for map {valid_nozzle_map_key} using tip type {active_tip_type}."
    )



from opentrons.types import Location, Mount, AxisMapType
from .helpers import stringify_location
from . import types as command_types
from typing import Optional


def move_to(
    mount: Mount, location: Location, speed: Optional[float]
) -> command_types.RobotMoveToCommand:
    location_text = stringify_location(location)
    text = f"Moving to {location_text} at {speed}"
    return {
        "name": command_types.ROBOT_MOVE_TO,
        "payload": {"mount": mount, "location": location, "text": text},
    }


def move_axis_to(
    axis_map: AxisMapType, speed: Optional[float]
) -> command_types.RobotMoveAxisToCommand:
    text = f"Moving to the provided absolute axis map {axis_map} at {speed}."
    return {
        "name": command_types.ROBOT_MOVE_AXES_TO,
        "payload": {"absolute_axes": axis_map, "text": text},
    }


def move_axis_relative(
    axis_map: AxisMapType, speed: Optional[float]
) -> command_types.RobotMoveAxisRelativeCommand:
    text = f"Moving to the provided relative axis map {axis_map} as {speed}"
    return {
        "name": command_types.ROBOT_MOVE_RELATIVE_TO,
        "payload": {"relative_axes": axis_map, "text": text},
    }


def open_gripper() -> command_types.RobotOpenGripperJawCommand:
    text = "Opening the gripper jaw."
    return {
        "name": command_types.ROBOT_OPEN_GRIPPER_JAW,
        "payload": {"text": text},
    }


def close_gripper(force: Optional[float]) -> command_types.RobotCloseGripperJawCommand:
    text = f"Closing the gripper jaw with force {force}."
    return {
        "name": command_types.ROBOT_CLOSE_GRIPPER_JAW,
        "payload": {"text": text},
    }



from typing import Dict, List, Optional, Tuple
from opentrons.util.async_helpers import ensure_yield

from opentrons.drivers.types import (
    ABSMeasurementMode,
    AbsorbanceReaderLidStatus,
    AbsorbanceReaderDeviceState,
    AbsorbanceReaderPlatePresence,
)

from .abstract import AbstractAbsorbanceReaderDriver


class SimulatingDriver(AbstractAbsorbanceReaderDriver):
    def __init__(
        self, model: Optional[str] = None, serial_number: Optional[str] = None
    ) -> None:
        self._lid_status = AbsorbanceReaderLidStatus.ON
        self._model = model if model else "absorbanceReaderV1"
        self._serial_number = serial_number

    def model(self) -> str:
        return self._model

    @ensure_yield
    async def get_device_info(self) -> Dict[str, str]:
        """Get device info"""
        return {
            "serial": self._serial_number if self._serial_number else "dummySerialAR",
            "model": "dummyModelAR",
            "version": "dummyVersionAR",
        }

    @ensure_yield
    async def connect(self) -> None:
        """Connect to absorbance reader"""
        pass

    @ensure_yield
    async def disconnect(self) -> None:
        """Disconnect from absorbance reader"""
        pass

    @ensure_yield
    async def is_connected(self) -> bool:
        """Check connection to absorbance reader"""
        return True

    @ensure_yield
    async def get_lid_status(self) -> AbsorbanceReaderLidStatus:
        return self._lid_status

    @ensure_yield
    async def get_available_wavelengths(self) -> List[int]:
        return [450, 570, 600, 650]

    @ensure_yield
    async def get_measurement(self) -> List[List[float]]:
        return [[0.0]]

    @ensure_yield
    async def initialize_measurement(
        self,
        wavelengths: List[int],
        mode: ABSMeasurementMode = ABSMeasurementMode.SINGLE,
        reference_wavelength: Optional[int] = None,
    ) -> None:
        pass

    @ensure_yield
    async def get_uptime(self) -> int:
        return 10

    @ensure_yield
    async def update_firmware(self, firmware_file_path: str) -> Tuple[bool, str]:
        return True, ""

    @ensure_yield
    async def get_status(self) -> AbsorbanceReaderDeviceState:
        return AbsorbanceReaderDeviceState.OK

    @ensure_yield
    async def get_plate_presence(self) -> AbsorbanceReaderPlatePresence:
        return AbsorbanceReaderPlatePresence.PRESENT



"""Protocol engine types for labware offset vectors.

This is a separate module to avoid circular imports.
"""
from __future__ import annotations
from typing import Any

from pydantic import BaseModel


# TODO(mm, 2022-11-07): Deduplicate with Vec3f.
class LabwareOffsetVector(BaseModel):
    """Offset, in deck coordinates from nominal to actual position."""

    x: float
    y: float
    z: float

    def __add__(self, other: Any) -> LabwareOffsetVector:
        """Adds two vectors together."""
        if not isinstance(other, LabwareOffsetVector):
            return NotImplemented
        return LabwareOffsetVector(
            x=self.x + other.x, y=self.y + other.y, z=self.z + other.z
        )

    def __sub__(self, other: Any) -> LabwareOffsetVector:
        """Subtracts two vectors."""
        if not isinstance(other, LabwareOffsetVector):
            return NotImplemented
        return LabwareOffsetVector(
            x=self.x - other.x, y=self.y - other.y, z=self.z - other.z
        )



"""State store actions.

Actions can be passed to the ActionDispatcher, where they will trigger
reactions in objects that subscribe to the pipeline, like the StateStore.
"""
import dataclasses
from datetime import datetime
from enum import Enum
from typing import List, Optional, Union

from opentrons_shared_data.errors import EnumeratedError
from opentrons_shared_data.labware.labware_definition import LabwareDefinition

from opentrons.hardware_control.types import DoorState
from opentrons.hardware_control.modules import LiveData

from ..commands import (
    Command,
    CommandCreate,
    CommandDefinedErrorData,
)
from ..error_recovery_policy import ErrorRecoveryPolicy, ErrorRecoveryType
from ..notes.notes import CommandNote
from ..state.update_types import StateUpdate
from ..types import (
    LabwareOffsetCreateInternal,
    ModuleDefinition,
    Liquid,
    DeckConfigurationType,
)


@dataclasses.dataclass(frozen=True)
class PlayAction:
    """Start or resume processing commands in the engine."""

    requested_at: datetime


class PauseSource(str, Enum):
    """The source of a PauseAction.

    Attributes:
        CLIENT: the pause came externally, from the engine client.
        PROTOCOL: the pause came from the protocol itself.
    """

    CLIENT = "client"
    PROTOCOL = "protocol"


@dataclasses.dataclass(frozen=True)
class PauseAction:
    """Pause processing commands in the engine."""

    source: PauseSource


@dataclasses.dataclass(frozen=True)
class StopAction:
    """Request engine execution to stop soon."""

    from_estop: bool = False


@dataclasses.dataclass(frozen=True)
class ResumeFromRecoveryAction:
    """See `ProtocolEngine.resume_from_recovery()`."""

    state_update: StateUpdate


@dataclasses.dataclass(frozen=True)
class FinishErrorDetails:
    """Error details for the payload of a FinishAction or HardwareStoppedAction."""

    error: Exception
    error_id: str
    created_at: datetime


@dataclasses.dataclass(frozen=True)
class FinishAction:
    """Gracefully stop processing commands in the engine."""

    set_run_status: bool = True
    """Whether to set the engine status depending on `error_details`.

    If True, the engine status will be marked `succeeded` or `failed`, depending on `error_details`.
    If False, the engine status will be marked `stopped`.
    """

    error_details: Optional[FinishErrorDetails] = None
    """The fatal error that caused the run to fail."""


@dataclasses.dataclass(frozen=True)
class HardwareStoppedAction:
    """An action dispatched after hardware has been stopped for good, for this engine instance."""

    completed_at: datetime

    finish_error_details: Optional[FinishErrorDetails]
    """The error that happened while doing post-run finish steps (homing and dropping tips)."""


@dataclasses.dataclass(frozen=True)
class DoorChangeAction:
    """Handle events coming in from hardware control."""

    door_state: DoorState


@dataclasses.dataclass(frozen=True)
class QueueCommandAction:
    """Add a command request to the queue."""

    command_id: str
    created_at: datetime
    request: CommandCreate
    request_hash: Optional[str]
    failed_command_id: Optional[str] = None


@dataclasses.dataclass(frozen=True)
class RunCommandAction:
    """Mark a given command as running.

    At the time of dispatching this action, the command must be queued,
    and no other command may be running.
    """

    command_id: str
    started_at: datetime


@dataclasses.dataclass(frozen=True)
class SucceedCommandAction:
    """Mark a given command as succeeded.

    At the time of dispatching this action, the command must be running.
    """

    command: Command
    """The command in its new succeeded state."""

    state_update: StateUpdate = dataclasses.field(
        # todo(mm, 2024-08-26): This has a default only to make it easier to transition
        # old tests while https://opentrons.atlassian.net/browse/EXEC-639 is in
        # progress. Make this mandatory when that's completed.
        default_factory=StateUpdate
    )


@dataclasses.dataclass(frozen=True)
class FailCommandAction:
    """Mark a given command as failed.

    At the time of dispatching this action, the command must be running.
    """

    command_id: str
    """The command to fail."""

    error_id: str
    """An ID to assign to the command's error.

    Must be unique to this occurrence of the error.

    todo(mm, 2024-05-13): This is redundant with `error` when it's a defined error.
    """

    failed_at: datetime
    """When the command failed.

    todo(mm, 2024-05-13): This is redundant with `error` when it's a defined error.
    """

    error: Union[CommandDefinedErrorData, EnumeratedError]
    """The error that caused the command to fail.

    If it was a defined error, this should be the `DefinedErrorData` that the command
    returned.

    If it was an undefined error, this should be the underlying exception
    that caused the command to fail, represented as an `EnumeratedError`.
    """

    notes: List[CommandNote]
    """Overwrite the command's `.notes` with these."""

    type: ErrorRecoveryType
    """How this error should be handled in the context of the overall run."""

    # This is a quick hack so FailCommandAction handlers can get the params of the
    # command that failed. We probably want this to be a new "failure details"
    # object instead, similar to how succeeded commands can send a "private result"
    # to Protocol Engine internals.
    running_command: Command
    """The command to fail, in its prior `running` state."""


@dataclasses.dataclass(frozen=True)
class AddLabwareOffsetAction:
    """Add a labware offset, to apply to subsequent `LoadLabwareCommand`s."""

    labware_offset_id: str
    created_at: datetime
    request: LabwareOffsetCreateInternal


@dataclasses.dataclass(frozen=True)
class AddLabwareDefinitionAction:
    """Add a labware definition, to apply to subsequent `LoadLabwareCommand`s."""

    definition: LabwareDefinition


@dataclasses.dataclass(frozen=True)
class AddLiquidAction:
    """Add a liquid, to apply to subsequent `LoadLiquid`s."""

    liquid: Liquid


@dataclasses.dataclass(frozen=True)
class SetDeckConfigurationAction:
    """See `ProtocolEngine.set_deck_configuration()`."""

    deck_configuration: Optional[DeckConfigurationType]


@dataclasses.dataclass(frozen=True)
class AddAddressableAreaAction:
    """Add a single addressable area to state.

    This differs from the deck configuration in SetDeckConfigurationAction which
    sends over a mapping of cutout fixtures. This action will only load one addressable
    area and that should be pre-validated before being sent via the action.
    """

    addressable_area_name: str


@dataclasses.dataclass(frozen=True)
class AddModuleAction:
    """Add an attached module directly to state without a location."""

    module_id: str
    serial_number: str
    definition: ModuleDefinition
    module_live_data: LiveData


@dataclasses.dataclass(frozen=True)
class ResetTipsAction:
    """Reset the tip tracking state of a given tip rack."""

    labware_id: str


@dataclasses.dataclass(frozen=True)
class SetPipetteMovementSpeedAction:
    """Set the speed of a pipette's X/Y/Z movements. Does not affect plunger speed.

    None will use the hardware API's default.
    """

    pipette_id: str
    speed: Optional[float]


@dataclasses.dataclass(frozen=True)
class SetErrorRecoveryPolicyAction:
    """See `ProtocolEngine.set_error_recovery_policy()`."""

    error_recovery_policy: ErrorRecoveryPolicy


Action = Union[
    PlayAction,
    PauseAction,
    StopAction,
    ResumeFromRecoveryAction,
    FinishAction,
    HardwareStoppedAction,
    DoorChangeAction,
    QueueCommandAction,
    RunCommandAction,
    SucceedCommandAction,
    FailCommandAction,
    AddLabwareOffsetAction,
    AddLabwareDefinitionAction,
    AddModuleAction,
    SetDeckConfigurationAction,
    AddAddressableAreaAction,
    AddLiquidAction,
    ResetTipsAction,
    SetPipetteMovementSpeedAction,
    SetErrorRecoveryPolicyAction,
]



from dataclasses import dataclass
from numpy import interp
from typing import Optional, Dict, Sequence, Tuple, List, Union

from opentrons_shared_data.liquid_classes.liquid_class_definition import (
    TransferProperties as SharedDataTransferProperties,
    AspirateProperties as SharedDataAspirateProperties,
    SingleDispenseProperties as SharedDataSingleDispenseProperties,
    MultiDispenseProperties as SharedDataMultiDispenseProperties,
    TipPosition as SharedDataTipPosition,
    DelayProperties as SharedDataDelayProperties,
    DelayParams as SharedDataDelayParams,
    TouchTipProperties as SharedDataTouchTipProperties,
    LiquidClassTouchTipParams as SharedDataTouchTipParams,
    MixProperties as SharedDataMixProperties,
    MixParams as SharedDataMixParams,
    BlowoutProperties as SharedDataBlowoutProperties,
    BlowoutParams as SharedDataBlowoutParams,
    ByTipTypeSetting as SharedByTipTypeSetting,
    Submerge as SharedDataSubmerge,
    RetractAspirate as SharedDataRetractAspirate,
    RetractDispense as SharedDataRetractDispense,
    BlowoutLocation,
    PositionReference,
    Coordinate,
)
from . import validation


class LiquidHandlingPropertyByVolume:
    def __init__(self, by_volume_property: Sequence[Tuple[float, float]]) -> None:
        self._initial_properties_by_volume = by_volume_property
        self._properties_by_volume: Dict[float, float] = {
            float(volume): value for volume, value in by_volume_property
        }
        # Volumes need to be sorted for proper interpolation of non-defined volumes, and the
        # corresponding values need to be in the same order for them to be interpolated correctly
        self._sorted_volumes: Tuple[float, ...] = ()
        self._sorted_values: Tuple[float, ...] = ()
        self._sort_volume_and_values()

    def as_dict(self) -> Dict[float, float]:
        """Get a dictionary representation of all set volumes and values along with the default."""
        return self._properties_by_volume

    def as_list_of_tuples(self) -> List[Tuple[float, float]]:
        """Get as list of tuples."""
        return list(self._properties_by_volume.items())

    def get_for_volume(self, volume: float) -> float:
        """Get a value by volume for this property. Volumes not defined will be interpolated between set volumes."""
        validated_volume = validation.ensure_positive_float(volume)
        if len(self._properties_by_volume) == 0:
            raise ValueError(
                "No properties found for any volumes. Cannot interpolate for the given volume."
            )
        try:
            return self._properties_by_volume[validated_volume]
        except KeyError:
            # If volume is not defined in dictionary, do a piecewise interpolation with existing sorted values
            return float(
                interp(validated_volume, self._sorted_volumes, self._sorted_values)
            )

    def set_for_all_volumes(self, value: float) -> None:
        """Override all existing volume-dependent values with the given value."""
        self.clear_values()
        self.set_for_volume(0, value)

    def set_for_volume(self, volume: float, value: float) -> None:
        """Add a new volume and value for the property for the interpolation curve."""
        validated_volume = validation.ensure_positive_float(volume)
        self._properties_by_volume[validated_volume] = value
        self._sort_volume_and_values()

    def delete_for_volume(self, volume: float) -> None:
        """Remove an existing volume and value from the property."""
        try:
            del self._properties_by_volume[volume]
        except KeyError:
            raise KeyError(f"No value set for volume {volume} uL")
        self._sort_volume_and_values()

    def clear_values(self) -> None:
        """Removes all existing volume and value pairs from the curve."""
        self._properties_by_volume = {}

    def reset_values(self) -> None:
        """Resets volumes and values to the default."""
        self._properties_by_volume = {
            float(volume): value for volume, value in self._initial_properties_by_volume
        }
        self._sort_volume_and_values()

    def _sort_volume_and_values(self) -> None:
        """Sort volume in increasing order along with corresponding values in matching order."""
        self._sorted_volumes, self._sorted_values = (
            zip(*sorted(self._properties_by_volume.items()))
            if len(self._properties_by_volume) > 0
            else [(), ()]
        )


# We use slots for this dataclass (and the rest of liquid properties) to prevent dynamic creation of attributes
# not defined in the class, not for any performance reasons. This is so that mistyping properties when overriding
# values will cause the protocol to fail analysis, rather than silently passing.
@dataclass(slots=True)
class TipPosition:

    _position_reference: PositionReference
    _offset: Coordinate

    @property
    def position_reference(self) -> PositionReference:
        return self._position_reference

    @position_reference.setter
    def position_reference(self, new_position: Union[str, PositionReference]) -> None:
        self._position_reference = (
            new_position
            if isinstance(new_position, PositionReference)
            else PositionReference(new_position)
        )

    @property
    def offset(self) -> Coordinate:
        return self._offset

    @offset.setter
    def offset(self, new_offset: Union[Sequence[float], Coordinate]) -> None:
        if isinstance(new_offset, Coordinate):
            new_coordinate: Sequence[Union[int, float]] = [
                new_offset.x,
                new_offset.y,
                new_offset.z,
            ]
        else:
            new_coordinate = new_offset
        x, y, z = validation.validate_coordinates(new_coordinate)
        self._offset = Coordinate(x=x, y=y, z=z)

    def as_shared_data_model(self) -> SharedDataTipPosition:
        return SharedDataTipPosition(
            positionReference=self._position_reference,
            offset=self.offset,
        )


@dataclass(slots=True)
class DelayProperties:

    _enabled: bool
    _duration: Optional[float]

    @property
    def enabled(self) -> bool:
        return self._enabled

    @enabled.setter
    def enabled(self, enable: bool) -> None:
        validated_enable = validation.ensure_boolean(enable)
        if validated_enable and self._duration is None:
            raise ValueError("duration must be set before enabling delay.")
        self._enabled = validated_enable

    @property
    def duration(self) -> Optional[float]:
        return self._duration

    @duration.setter
    def duration(self, new_duration: float) -> None:
        validated_duration = validation.ensure_positive_float(new_duration)
        self._duration = validated_duration

    def as_shared_data_model(self) -> SharedDataDelayProperties:
        return SharedDataDelayProperties(
            enable=self._enabled,
            params=SharedDataDelayParams(duration=self.duration)
            if self.duration is not None
            else None,
        )


@dataclass(slots=True)
class TouchTipProperties:

    _enabled: bool
    _z_offset: Optional[float]
    _mm_from_edge: Optional[float]
    _speed: Optional[float]

    @property
    def enabled(self) -> bool:
        return self._enabled

    @enabled.setter
    def enabled(self, enable: bool) -> None:
        validated_enable = validation.ensure_boolean(enable)
        if validated_enable and (
            self._z_offset is None or self._mm_from_edge is None or self._speed is None
        ):
            raise ValueError(
                "z_offset, mm_from_edge and speed must be set before enabling touch tip."
            )
        self._enabled = validated_enable

    @property
    def z_offset(self) -> Optional[float]:
        return self._z_offset

    @z_offset.setter
    def z_offset(self, new_offset: float) -> None:
        validated_offset = validation.ensure_float(new_offset)
        self._z_offset = validated_offset

    @property
    def mm_from_edge(self) -> Optional[float]:
        return self._mm_from_edge

    @mm_from_edge.setter
    def mm_from_edge(self, new_mm: float) -> None:
        validated_mm = validation.ensure_float(new_mm)
        self._mm_from_edge = validated_mm

    @property
    def speed(self) -> Optional[float]:
        return self._speed

    @speed.setter
    def speed(self, new_speed: float) -> None:
        validated_speed = validation.ensure_greater_than_zero_float(new_speed)
        self._speed = validated_speed

    def _get_shared_data_params(self) -> Optional[SharedDataTouchTipParams]:
        """Get the touch tip params in schema v1 shape."""
        if (
            self._z_offset is not None
            and self._mm_from_edge is not None
            and self._speed is not None
        ):
            return SharedDataTouchTipParams(
                zOffset=self._z_offset,
                mmFromEdge=self._mm_from_edge,
                speed=self._speed,
            )
        else:
            return None

    def as_shared_data_model(self) -> SharedDataTouchTipProperties:
        return SharedDataTouchTipProperties(
            enable=self._enabled,
            params=self._get_shared_data_params(),
        )


@dataclass(slots=True)
class MixProperties:

    _enabled: bool
    _repetitions: Optional[int]
    _volume: Optional[float]

    @property
    def enabled(self) -> bool:
        return self._enabled

    @enabled.setter
    def enabled(self, enable: bool) -> None:
        validated_enable = validation.ensure_boolean(enable)
        if validated_enable and (self._repetitions is None or self._volume is None):
            raise ValueError("repetitions and volume must be set before enabling mix.")
        self._enabled = validated_enable

    @property
    def repetitions(self) -> Optional[int]:
        return self._repetitions

    @repetitions.setter
    def repetitions(self, new_repetitions: int) -> None:
        validated_repetitions = validation.ensure_positive_int(new_repetitions)
        self._repetitions = validated_repetitions

    @property
    def volume(self) -> Optional[float]:
        return self._volume

    @volume.setter
    def volume(self, new_volume: float) -> None:
        validated_volume = validation.ensure_greater_than_zero_float(new_volume)
        self._volume = validated_volume

    def _get_shared_data_params(self) -> Optional[SharedDataMixParams]:
        """Get the mix params in schema v1 shape."""
        if self._repetitions is not None and self._volume is not None:
            return SharedDataMixParams(
                repetitions=self._repetitions,
                volume=self._volume,
            )
        else:
            return None

    def as_shared_data_model(self) -> SharedDataMixProperties:
        return SharedDataMixProperties(
            enable=self._enabled,
            params=self._get_shared_data_params(),
        )


@dataclass(slots=True)
class BlowoutProperties:

    _enabled: bool
    _location: Optional[BlowoutLocation]
    _flow_rate: Optional[float]

    @property
    def enabled(self) -> bool:
        return self._enabled

    @enabled.setter
    def enabled(self, enable: bool) -> None:
        validated_enable = validation.ensure_boolean(enable)
        if validated_enable and (self._location is None or self._flow_rate is None):
            raise ValueError(
                "location and flow_rate must be set before enabling blowout."
            )
        self._enabled = validated_enable

    @property
    def location(self) -> Optional[BlowoutLocation]:
        return self._location

    @location.setter
    def location(self, new_location: str) -> None:
        self._location = BlowoutLocation(new_location)

    @property
    def flow_rate(self) -> Optional[float]:
        return self._flow_rate

    @flow_rate.setter
    def flow_rate(self, new_flow_rate: float) -> None:
        validated_flow_rate = validation.ensure_greater_than_zero_float(new_flow_rate)
        self._flow_rate = validated_flow_rate

    def _get_shared_data_params(self) -> Optional[SharedDataBlowoutParams]:
        """Get the mix params in schema v1 shape."""
        if self._location is not None and self._flow_rate is not None:
            return SharedDataBlowoutParams(
                location=self._location,
                flowRate=self._flow_rate,
            )
        else:
            return None

    def as_shared_data_model(self) -> SharedDataBlowoutProperties:
        return SharedDataBlowoutProperties(
            enable=self._enabled,
            params=self._get_shared_data_params(),
        )


@dataclass(slots=True)
class _SubmergeRetractCommon:

    _speed: float
    _delay: DelayProperties

    @property
    def speed(self) -> float:
        return self._speed

    @speed.setter
    def speed(self, new_speed: float) -> None:
        validated_speed = validation.ensure_positive_float(new_speed)
        self._speed = validated_speed

    @property
    def delay(self) -> DelayProperties:
        return self._delay


@dataclass(slots=True)
class Submerge(_SubmergeRetractCommon):

    _start_position: TipPosition

    @property
    def start_position(self) -> TipPosition:
        return self._start_position

    def as_shared_data_model(self) -> SharedDataSubmerge:
        return SharedDataSubmerge(
            startPosition=self._start_position.as_shared_data_model(),
            speed=self._speed,
            delay=self._delay.as_shared_data_model(),
        )


@dataclass(slots=True)
class RetractAspirate(_SubmergeRetractCommon):

    _end_position: TipPosition
    _air_gap_by_volume: LiquidHandlingPropertyByVolume
    _touch_tip: TouchTipProperties

    @property
    def end_position(self) -> TipPosition:
        return self._end_position

    @property
    def air_gap_by_volume(self) -> LiquidHandlingPropertyByVolume:
        return self._air_gap_by_volume

    @property
    def touch_tip(self) -> TouchTipProperties:
        return self._touch_tip

    def as_shared_data_model(self) -> SharedDataRetractAspirate:
        return SharedDataRetractAspirate(
            endPosition=self._end_position.as_shared_data_model(),
            speed=self._speed,
            airGapByVolume=self._air_gap_by_volume.as_list_of_tuples(),
            touchTip=self._touch_tip.as_shared_data_model(),
            delay=self._delay.as_shared_data_model(),
        )


@dataclass(slots=True)
class RetractDispense(_SubmergeRetractCommon):
    _end_position: TipPosition
    _air_gap_by_volume: LiquidHandlingPropertyByVolume
    _touch_tip: TouchTipProperties
    _blowout: BlowoutProperties

    @property
    def end_position(self) -> TipPosition:
        return self._end_position

    @property
    def air_gap_by_volume(self) -> LiquidHandlingPropertyByVolume:
        return self._air_gap_by_volume

    @property
    def touch_tip(self) -> TouchTipProperties:
        return self._touch_tip

    @property
    def blowout(self) -> BlowoutProperties:
        return self._blowout

    def as_shared_data_model(self) -> SharedDataRetractDispense:
        return SharedDataRetractDispense(
            endPosition=self._end_position.as_shared_data_model(),
            speed=self._speed,
            airGapByVolume=self._air_gap_by_volume.as_list_of_tuples(),
            blowout=self._blowout.as_shared_data_model(),
            touchTip=self._touch_tip.as_shared_data_model(),
            delay=self._delay.as_shared_data_model(),
        )


@dataclass(slots=True)
class _BaseLiquidHandlingProperties:

    _submerge: Submerge
    _flow_rate_by_volume: LiquidHandlingPropertyByVolume
    _correction_by_volume: LiquidHandlingPropertyByVolume
    _delay: DelayProperties

    @property
    def submerge(self) -> Submerge:
        return self._submerge

    @property
    def flow_rate_by_volume(self) -> LiquidHandlingPropertyByVolume:
        return self._flow_rate_by_volume

    @property
    def correction_by_volume(self) -> LiquidHandlingPropertyByVolume:
        return self._correction_by_volume

    @property
    def delay(self) -> DelayProperties:
        return self._delay


@dataclass(slots=True)
class AspirateProperties(_BaseLiquidHandlingProperties):

    _aspirate_position: TipPosition
    _retract: RetractAspirate
    _pre_wet: bool
    _mix: MixProperties

    @property
    def aspirate_position(self) -> TipPosition:
        return self._aspirate_position

    @property
    def pre_wet(self) -> bool:
        return self._pre_wet

    @pre_wet.setter
    def pre_wet(self, new_setting: bool) -> None:
        validated_setting = validation.ensure_boolean(new_setting)
        self._pre_wet = validated_setting

    @property
    def retract(self) -> RetractAspirate:
        return self._retract

    @property
    def mix(self) -> MixProperties:
        return self._mix

    def as_shared_data_model(self) -> SharedDataAspirateProperties:
        return SharedDataAspirateProperties(
            submerge=self._submerge.as_shared_data_model(),
            retract=self._retract.as_shared_data_model(),
            aspiratePosition=self._aspirate_position.as_shared_data_model(),
            flowRateByVolume=self._flow_rate_by_volume.as_list_of_tuples(),
            preWet=self._pre_wet,
            mix=self._mix.as_shared_data_model(),
            delay=self._delay.as_shared_data_model(),
            correctionByVolume=self._correction_by_volume.as_list_of_tuples(),
        )


@dataclass(slots=True)
class SingleDispenseProperties(_BaseLiquidHandlingProperties):

    _dispense_position: TipPosition
    _retract: RetractDispense
    _push_out_by_volume: LiquidHandlingPropertyByVolume
    _mix: MixProperties

    @property
    def dispense_position(self) -> TipPosition:
        return self._dispense_position

    @property
    def push_out_by_volume(self) -> LiquidHandlingPropertyByVolume:
        return self._push_out_by_volume

    @property
    def retract(self) -> RetractDispense:
        return self._retract

    @property
    def mix(self) -> MixProperties:
        return self._mix

    def as_shared_data_model(self) -> SharedDataSingleDispenseProperties:
        return SharedDataSingleDispenseProperties(
            submerge=self._submerge.as_shared_data_model(),
            retract=self._retract.as_shared_data_model(),
            dispensePosition=self._dispense_position.as_shared_data_model(),
            flowRateByVolume=self._flow_rate_by_volume.as_list_of_tuples(),
            mix=self._mix.as_shared_data_model(),
            pushOutByVolume=self._push_out_by_volume.as_list_of_tuples(),
            delay=self._delay.as_shared_data_model(),
            correctionByVolume=self._correction_by_volume.as_list_of_tuples(),
        )


@dataclass(slots=True)
class MultiDispenseProperties(_BaseLiquidHandlingProperties):

    _dispense_position: TipPosition
    _retract: RetractDispense
    _conditioning_by_volume: LiquidHandlingPropertyByVolume
    _disposal_by_volume: LiquidHandlingPropertyByVolume

    @property
    def dispense_position(self) -> TipPosition:
        return self._dispense_position

    @property
    def retract(self) -> RetractDispense:
        return self._retract

    @property
    def conditioning_by_volume(self) -> LiquidHandlingPropertyByVolume:
        return self._conditioning_by_volume

    @property
    def disposal_by_volume(self) -> LiquidHandlingPropertyByVolume:
        return self._disposal_by_volume

    def as_shared_data_model(self) -> SharedDataMultiDispenseProperties:
        return SharedDataMultiDispenseProperties(
            submerge=self._submerge.as_shared_data_model(),
            retract=self._retract.as_shared_data_model(),
            dispensePosition=self._dispense_position.as_shared_data_model(),
            flowRateByVolume=self._flow_rate_by_volume.as_list_of_tuples(),
            conditioningByVolume=self._conditioning_by_volume.as_list_of_tuples(),
            disposalByVolume=self._disposal_by_volume.as_list_of_tuples(),
            delay=self._delay.as_shared_data_model(),
            correctionByVolume=self._correction_by_volume.as_list_of_tuples(),
        )


@dataclass(slots=True)
class TransferProperties:
    _aspirate: AspirateProperties
    _dispense: SingleDispenseProperties
    _multi_dispense: Optional[MultiDispenseProperties]

    @property
    def aspirate(self) -> AspirateProperties:
        """Aspirate properties."""
        return self._aspirate

    @property
    def dispense(self) -> SingleDispenseProperties:
        """Single dispense properties."""
        return self._dispense

    @property
    def multi_dispense(self) -> Optional[MultiDispenseProperties]:
        """Multi dispense properties."""
        return self._multi_dispense


def _build_tip_position(tip_position: SharedDataTipPosition) -> TipPosition:
    return TipPosition(
        _position_reference=tip_position.positionReference, _offset=tip_position.offset
    )


def _build_delay_properties(
    delay_properties: SharedDataDelayProperties,
) -> DelayProperties:
    if delay_properties.params is not None:
        duration = delay_properties.params.duration
    else:
        duration = None
    return DelayProperties(_enabled=delay_properties.enable, _duration=duration)


def _build_touch_tip_properties(
    touch_tip_properties: SharedDataTouchTipProperties,
) -> TouchTipProperties:
    if touch_tip_properties.params is not None:
        z_offset = touch_tip_properties.params.zOffset
        mm_from_edge = touch_tip_properties.params.mmFromEdge
        speed = touch_tip_properties.params.speed
    else:
        z_offset = None
        mm_from_edge = None
        speed = None
    return TouchTipProperties(
        _enabled=touch_tip_properties.enable,
        _z_offset=z_offset,
        _mm_from_edge=mm_from_edge,
        _speed=speed,
    )


def _build_mix_properties(
    mix_properties: SharedDataMixProperties,
) -> MixProperties:
    if mix_properties.params is not None:
        repetitions = mix_properties.params.repetitions
        volume = mix_properties.params.volume
    else:
        repetitions = None
        volume = None
    return MixProperties(
        _enabled=mix_properties.enable, _repetitions=repetitions, _volume=volume
    )


def _build_blowout_properties(
    blowout_properties: SharedDataBlowoutProperties,
) -> BlowoutProperties:
    if blowout_properties.params is not None:
        location = blowout_properties.params.location
        flow_rate = blowout_properties.params.flowRate
    else:
        location = None
        flow_rate = None
    return BlowoutProperties(
        _enabled=blowout_properties.enable, _location=location, _flow_rate=flow_rate
    )


def _build_submerge(
    submerge_properties: SharedDataSubmerge,
) -> Submerge:
    return Submerge(
        _start_position=_build_tip_position(submerge_properties.startPosition),
        _speed=submerge_properties.speed,
        _delay=_build_delay_properties(submerge_properties.delay),
    )


def _build_retract_aspirate(
    retract_aspirate: SharedDataRetractAspirate,
) -> RetractAspirate:
    return RetractAspirate(
        _end_position=_build_tip_position(retract_aspirate.endPosition),
        _speed=retract_aspirate.speed,
        _air_gap_by_volume=LiquidHandlingPropertyByVolume(
            retract_aspirate.airGapByVolume
        ),
        _touch_tip=_build_touch_tip_properties(retract_aspirate.touchTip),
        _delay=_build_delay_properties(retract_aspirate.delay),
    )


def _build_retract_dispense(
    retract_dispense: SharedDataRetractDispense,
) -> RetractDispense:
    return RetractDispense(
        _end_position=_build_tip_position(retract_dispense.endPosition),
        _speed=retract_dispense.speed,
        _air_gap_by_volume=LiquidHandlingPropertyByVolume(
            retract_dispense.airGapByVolume
        ),
        _blowout=_build_blowout_properties(retract_dispense.blowout),
        _touch_tip=_build_touch_tip_properties(retract_dispense.touchTip),
        _delay=_build_delay_properties(retract_dispense.delay),
    )


def build_aspirate_properties(
    aspirate_properties: SharedDataAspirateProperties,
) -> AspirateProperties:
    return AspirateProperties(
        _submerge=_build_submerge(aspirate_properties.submerge),
        _retract=_build_retract_aspirate(aspirate_properties.retract),
        _aspirate_position=_build_tip_position(aspirate_properties.aspiratePosition),
        _flow_rate_by_volume=LiquidHandlingPropertyByVolume(
            aspirate_properties.flowRateByVolume
        ),
        _correction_by_volume=LiquidHandlingPropertyByVolume(
            aspirate_properties.correctionByVolume
        ),
        _pre_wet=aspirate_properties.preWet,
        _mix=_build_mix_properties(aspirate_properties.mix),
        _delay=_build_delay_properties(aspirate_properties.delay),
    )


def build_single_dispense_properties(
    single_dispense_properties: SharedDataSingleDispenseProperties,
) -> SingleDispenseProperties:
    return SingleDispenseProperties(
        _submerge=_build_submerge(single_dispense_properties.submerge),
        _retract=_build_retract_dispense(single_dispense_properties.retract),
        _dispense_position=_build_tip_position(
            single_dispense_properties.dispensePosition
        ),
        _flow_rate_by_volume=LiquidHandlingPropertyByVolume(
            single_dispense_properties.flowRateByVolume
        ),
        _correction_by_volume=LiquidHandlingPropertyByVolume(
            single_dispense_properties.correctionByVolume
        ),
        _mix=_build_mix_properties(single_dispense_properties.mix),
        _push_out_by_volume=LiquidHandlingPropertyByVolume(
            single_dispense_properties.pushOutByVolume
        ),
        _delay=_build_delay_properties(single_dispense_properties.delay),
    )


def build_multi_dispense_properties(
    multi_dispense_properties: Optional[SharedDataMultiDispenseProperties],
) -> Optional[MultiDispenseProperties]:
    if multi_dispense_properties is None:
        return None
    return MultiDispenseProperties(
        _submerge=_build_submerge(multi_dispense_properties.submerge),
        _retract=_build_retract_dispense(multi_dispense_properties.retract),
        _dispense_position=_build_tip_position(
            multi_dispense_properties.dispensePosition
        ),
        _flow_rate_by_volume=LiquidHandlingPropertyByVolume(
            multi_dispense_properties.flowRateByVolume
        ),
        _correction_by_volume=LiquidHandlingPropertyByVolume(
            multi_dispense_properties.correctionByVolume
        ),
        _conditioning_by_volume=LiquidHandlingPropertyByVolume(
            multi_dispense_properties.conditioningByVolume
        ),
        _disposal_by_volume=LiquidHandlingPropertyByVolume(
            multi_dispense_properties.disposalByVolume
        ),
        _delay=_build_delay_properties(multi_dispense_properties.delay),
    )


def build_transfer_properties(
    transfer_properties: Union[SharedDataTransferProperties, SharedByTipTypeSetting],
) -> TransferProperties:
    if isinstance(transfer_properties, SharedByTipTypeSetting):
        _transfer_properties = SharedDataTransferProperties(
            aspirate=transfer_properties.aspirate,
            singleDispense=transfer_properties.singleDispense,
            multiDispense=transfer_properties.multiDispense,
        )
    else:
        _transfer_properties = transfer_properties
    return TransferProperties(
        _aspirate=build_aspirate_properties(_transfer_properties.aspirate),
        _dispense=build_single_dispense_properties(_transfer_properties.singleDispense),
        _multi_dispense=build_multi_dispense_properties(
            _transfer_properties.multiDispense
        ),
    )
